
// ********************************************************
// simulation logging

#define _SAMPLE_LOG_LEVEL             (1 << 0)
#define _SEGMENT_LOG_LEVEL            (1 << 1)
#define _PATH_LOG_LEVEL               (1 << 2)
#define _SIMULATION_LOG_LEVEL         (1 << 3)
#define _WARNING_LOG_LEVEL            (1 << 4)
#define _ERROR_LOG_LEVEL              (1 << 5)
#define _NO_ERROR_TERMINATE_LOG_LEVEL (1 << 6)
#define _ALL_LOG_LEVELS        _SAMPLE_LOG_LEVEL|_SEGMENT_LOG_LEVEL|_PATH_LOG_LEVEL|_SIMULATION_LOG_LEVEL|_WARNING_LOG_LEVEL|_ERROR_LOG_LEVEL|_NO_ERROR_TERMINATE_LOG_LEVEL
#define _LOG_LEVELS_BEING_LOGGED (_PATH_LOG_LEVEL|_SIMULATION_LOG_LEVEL|_WARNING_LOG_LEVEL|_ERROR_LOG_LEVEL|_NO_ERROR_TERMINATE_LOG_LEVEL)

#define real Re
#define imag Im

#include <complex>

#undef real
#undef imag


#include <stdio.h>

#define _LOG(logLevel, ...) \
  do { \
    if (logLevel & _LOG_LEVELS_BEING_LOGGED) { \
      if ((_rank == 0) || (logLevel & ~(_SIMULATION_LOG_LEVEL | _SEGMENT_LOG_LEVEL | _SAMPLE_LOG_LEVEL | _NO_ERROR_TERMINATE_LOG_LEVEL))) { \
        if (logLevel & ~(_SIMULATION_LOG_LEVEL | _SEGMENT_LOG_LEVEL | _SAMPLE_LOG_LEVEL | _NO_ERROR_TERMINATE_LOG_LEVEL)) \
          printf("Rank[%i]: ", _rank); \
        if (logLevel & (_ERROR_LOG_LEVEL | _WARNING_LOG_LEVEL)) \
            printf("%s:%i: ", __FILE__, __LINE__); \
        printf(__VA_ARGS__); \
        fflush(stdout); \
      } \
      if (logLevel & _ERROR_LOG_LEVEL) \
        MPI_Abort(MPI_COMM_WORLD, 1); \
      else if (logLevel & _NO_ERROR_TERMINATE_LOG_LEVEL) \
        MPI_Finalize(); \
      if (logLevel & (_ERROR_LOG_LEVEL | _NO_ERROR_TERMINATE_LOG_LEVEL)) \
        exit(logLevel == _ERROR_LOG_LEVEL); \
    } \
  } while (0)

// ********************************************************
// simulation includes

#include <xpdeint_platform.h>
#include <cmath>
#include <string>
#include <cstring>
#include <fstream>
#include <sstream>
#include <cstdlib>

#if CFG_OSAPI == CFG_OSAPI_POSIX // These are POSIX headers (i.e. non-windows)
  #include <sys/time.h>
#endif // POSIX

#ifdef __APPLE__
  #include <Availability.h>
  #if __MAC_OS_X_VERSION_MIN_REQUIRED >= 1080
    #define OS_OBJECT_USE_OBJC 0 // Don't make dispatch and xpc objects Objective-C objects.
    #include <IOKit/pwr_mgt/IOPMLib.h> // To disable user idle sleep on Mountain Lion
  #endif
#endif

#include <time.h>
#include <list>
#include <vector>
#include <algorithm>

#include <utility>
#include <map>

#if (CFG_COMPILER == CFG_COMPILER_MSVC)
  #define FFTW_DLL
#endif

#include <fftw3.h>
#include <sys/stat.h>
#include <sys/types.h>

#define _xmds_malloc fftw_malloc
#define xmds_free fftw_free

#include <mpi.h>

#define H5_USE_16_API
#include <hdf5.h>

#if !defined(HAVE_H5LEXISTS)
htri_t H5Lexists(hid_t loc_id, const char *name, hid_t lapl_id)
{
  H5E_auto_t error_func;
  void* error_client_data;
  // Squelch errors generated by H5Gget_objinfo. It will report errors when it can't find an object
  // but that's the purpose of calling it.
  H5Eget_auto(&error_func, &error_client_data);
  H5Eset_auto(NULL, NULL);
  herr_t err = H5Gget_objinfo(loc_id, name, false, NULL);
  H5Eset_auto(error_func, error_client_data);
  if (err >= 0)
    return true;
  else
    return false;
}
#endif

#define H5T_NATIVE_REAL H5T_NATIVE_DOUBLE
#if defined(HAVE_HDF5_HL)
  #include <hdf5_hl.h>
#endif


typedef long integer;
typedef double real;
typedef std::complex<real> XMDSComplexType;

#include <xpdeint.h>

#define complex XMDSComplexType

const complex i(0.0, 1.0);

using namespace std;

#if CFG_COMPILER == CFG_COMPILER_ICC
  //
  // Disable ICC's warning: label was declared but never referenced
  //
  #pragma warning ( disable : 177 )
#endif

inline void *xmds_malloc(size_t size);

// ********************************************************
// DEFINES
// ********************************************************

// ********************************************************
//   Simulation defines
#define _EPSILON 1e-6
#ifndef INFINITY
#define INFINITY HUGE_VAL
#endif

#ifndef MAX
#define MAX(a, b) \
  ({ typeof(a) _a = (a); \
     typeof(b) _b = (b); \
     _a > _b ? _a : _b; })
#endif

#ifndef MIN
#define MIN(a, b) \
   ({ typeof(a) _a = (a); \
      typeof(b) _b = (b); \
      _a < _b ? _a : _b; })
#endif


// ********************************************************
//   Auto-vectorisation defines

#define _MAKE_AUTOVEC_VARIABLE(x) real* const __restrict__ x ## _autovec = (real*) x
#define _AUTOVEC(x) (x ## _autovec)

// ********************************************************
//   MPI Multipath Simulation Driver defines
#undef MPI_REAL
#define MPI_REAL MPI_DOUBLE
#define _n_paths 1

// ********************************************************
//   Geometry defines
#define _lattice_n ((int)21)
#define _min_n     ((long)0)
#define _max_n     ((long)20)
#define _dn        ((long)1)

#define _lattice_x ((int)256)
#define _min_x     ((real)-8)
#define _max_x     ((real)8)
#define _dx        ((real)((_max_x - _min_x)/_lattice_x))

#define _lattice_kx ((int)256)
#define _dkx        (2.0*M_PI/(_max_x - _min_x))
#define _min_kx     (-(_lattice_kx/2) * _dkx)
#define _max_kx     ((_lattice_kx - 1)/2 * _dkx)

#define _lattice_nprime ((int)21)
#define _min_nprime     ((long)0)
#define _max_nprime     ((long)20)
#define _dnprime        ((long)1)

// ********************************************************
//   field n defines
#define _n_ndims 1


// vector coefficients defines
#define _n_coefficients_ncomponents 2

// ********************************************************
//   field nx defines
#define _nx_ndims 2


// vector oscillator_states defines
#define _nx_oscillator_states_ncomponents 1

// ********************************************************
//   field x defines
#define _x_ndims 1


// vector densities defines
#define _x_densities_ncomponents 2

// ********************************************************
//   segment 1 (RK89 adaptive-step integrator) defines
// vector coefficients defines
#define _segment1_operator0_delta_a_field_coefficients_ncomponents 2

// ********************************************************
//   field mg0_sampling defines
#define _mg0_sampling_ndims 1


// ********************************************************
//   field mg0_output defines
#define _mg0_output_ndims 2


#define _mg0_output_lattice_t ((int)201)
#define _mg0_output_min_t     (_mg0_output_t[0])
#define _mg0_output_max_t     (_mg0_output_t[_mg0_output_lattice_t-1])
#define _mg0_output_dt        (_mg0_output_t[_index_t+1]-_mg0_output_t[_index_t])

// vector raw defines
#define _mg0_output_raw_ncomponents 6

// vector processed defines
#define _mg0_output_processed_ncomponents 6

// ********************************************************
//   field mg1_sampling defines
#define _mg1_sampling_ndims 1


// ********************************************************
//   field mg1_output defines
#define _mg1_output_ndims 1


#define _mg1_output_lattice_t ((int)201)
#define _mg1_output_min_t     (_mg1_output_t[0])
#define _mg1_output_max_t     (_mg1_output_t[_mg1_output_lattice_t-1])
#define _mg1_output_dt        (_mg1_output_t[_index_t+1]-_mg1_output_t[_index_t])

// vector raw defines
#define _mg1_output_raw_ncomponents 3

// vector processed defines
#define _mg1_output_processed_ncomponents 3

// ********************************************************
//   field mg2_sampling defines
#define _mg2_sampling_ndims 1


// ********************************************************
//   field mg2_output defines
#define _mg2_output_ndims 1


#define _mg2_output_lattice_t ((int)201)
#define _mg2_output_min_t     (_mg2_output_t[0])
#define _mg2_output_max_t     (_mg2_output_t[_mg2_output_lattice_t-1])
#define _mg2_output_dt        (_mg2_output_t[_index_t+1]-_mg2_output_t[_index_t])

// vector raw defines
#define _mg2_output_raw_ncomponents 1

// vector processed defines
#define _mg2_output_processed_ncomponents 1

// ********************************************************
//   field mg3_sampling defines
#define _mg3_sampling_ndims 1


// ********************************************************
//   field mg3_output defines
#define _mg3_output_ndims 2


#define _mg3_output_lattice_t ((int)201)
#define _mg3_output_min_t     (_mg3_output_t[0])
#define _mg3_output_max_t     (_mg3_output_t[_mg3_output_lattice_t-1])
#define _mg3_output_dt        (_mg3_output_t[_index_t+1]-_mg3_output_t[_index_t])

// vector raw defines
#define _mg3_output_raw_ncomponents 3

// vector processed defines
#define _mg3_output_processed_ncomponents 3

// ********************************************************
//   field segment1_operator0_looping_field defines
#define _segment1_operator0_looping_field_ndims 1


// ********************************************************
//   field segment1_operator0_delta_a_field defines
#define _segment1_operator0_delta_a_field_ndims 1



// ********************************************************
// GLOBALS
// ********************************************************


// ********************************************************
//   Simulation globals

string gsArgsAndValues = "";


// gPathID is used to keep track of the current path number when doing multipath simulations.
// It's needed for the export_all_paths option. Note his counter is per rank, so when
// running the mpi-multi-path driver it runs from 0 to _n_paths / _size - 1.
long gPathID = 0;    

  
real t;

// ********************************************************
//   Transform Multiplexer globals
typedef pair<ptrdiff_t, ptrdiff_t> _basis_pair;
typedef void (*transform_function)(bool, real, real* const __restrict__, real* const __restrict__, ptrdiff_t, ptrdiff_t);

// Less than operator needed by the C++ map class
struct _basis_pair_less_than
{
  bool operator()(const _basis_pair& _x, const _basis_pair& _y) const {
    return (_x.first < _y.first) || ((_x.first == _y.first) && (_x.second < _y.second));
  }
};

struct _transform_step
{
  transform_function _func;
  bool _forward;
  bool _out_of_place;
  ptrdiff_t _prefix_lattice;
  ptrdiff_t _postfix_lattice;
};

// Structure to hold the basis change information
struct _basis_transform_t
{
  vector<_transform_step> _transform_steps;
  real _multiplier;
  
  _basis_transform_t(real _multiplier_in = 1.0) : _multiplier(_multiplier_in) {}
  
  _basis_transform_t(const _basis_transform_t& _b) : _transform_steps(_b._transform_steps), _multiplier(_b._multiplier) {}
  
  void append(transform_function _func, bool _forward, bool _out_of_place, ptrdiff_t _prefix_lattice, ptrdiff_t _postfix_lattice)
  {
    _transform_steps.push_back((_transform_step){_func, _forward, _out_of_place, _prefix_lattice, _postfix_lattice});
  }
};

// Map type for holding (old_basis, new_basis) -> _basis_transform_t mappings
typedef map<_basis_pair, _basis_transform_t, _basis_pair_less_than> _basis_map;


real *_auxiliary_array = NULL;

const char *_basis_identifiers[] = {
};

// ********************************************************
//   'Globals' element globals

#line 32 "Assignment1_program_solutionb_v1.xmds"

// Parameters, some dimensionless units, where hbar=1, sigmaosc=1
const double omega0 = 0.0;
const double Delta0 = (2.0*M_PI)*0.3;

const double omega_i = (2.0*M_PI)*0.4;
const double chi_i = (2.0*M_PI)*0.2;
const double kappa_i_bar = (2.0*M_PI)*0.15;     

const double sigmaosc = 1.0;  

//initial state   
const double c1 = 1.0;                               //for (b)
const double c2 = 0.0;    

//***************************************************************  
// FUNCTION LIBRARY (room for own, non-xmds function definitions)      
//***************************************************************  
unsigned long factorial(unsigned long f){
   if (f) return(f * factorial(f - 1));
   return 1;
}

//***************************************************************
unsigned long integerpower(unsigned long argument,unsigned long  exponent){
unsigned long result=1;

for(unsigned long mm=0;mm<exponent;mm++)     
   result*=argument;

   return(result);    
}

//*************************************************************** 
double Hermite(double Rj,unsigned long degree){
  if(degree==0) 
    return 1.00000000000000000000;
 else if(degree==1)
    return 2.000000000000000*Rj;
 else{ 
    unsigned long alpha=degree-1;
    double value=2.000000000000000*(Rj*Hermite(Rj,alpha)-alpha*Hermite(Rj,alpha-1));
    return value;
 }      
}   


#line 440 "Assignment1_program_solutionb_v1.cc"

// ********************************************************
//   FFTW3 globals
const real _inverse_sqrt_2pi = 1.0 / sqrt(2.0 * M_PI); 
string _fftwWisdomPath;

// ********************************************************
//   MPI Multipath Simulation Driver globals
int _rank, _size;

// ********************************************************
//   Geometry globals
long* _n = NULL;

real* _x = NULL;

real* _kx = NULL;

long* _nprime = NULL;

// ********************************************************
//   field n globals
// vector coefficients globals
size_t _n_coefficients_alloc_size = 0;
complex* _n_coefficients = NULL;
complex* _active_n_coefficients = NULL;

// ********************************************************
//   field nx globals
// vector oscillator_states globals
size_t _nx_oscillator_states_alloc_size = 0;
real* _nx_oscillator_states = NULL;
real* _active_nx_oscillator_states = NULL;

// ********************************************************
//   field x globals
// vector densities globals
size_t _x_densities_alloc_size = 0;
complex* _x_densities = NULL;
complex* _active_x_densities = NULL;

// ********************************************************
//   segment 1 (RK89 adaptive-step integrator) globals
complex* _segment1_akafield_n_coefficients;
complex* _segment1_akbfield_n_coefficients;
complex* _segment1_akcfield_n_coefficients;
complex* _segment1_akdfield_n_coefficients;
complex* _segment1_akefield_n_coefficients;
complex* _segment1_akffield_n_coefficients;
complex* _segment1_akgfield_n_coefficients;
complex* _segment1_akhfield_n_coefficients;
complex* _segment1_akifield_n_coefficients;
complex* _segment1_akjfield_n_coefficients;
complex* _segment1_initial_n_coefficients;

// vector coefficients globals
size_t _segment1_operator0_delta_a_field_coefficients_alloc_size = 0;
complex* _segment1_operator0_delta_a_field_coefficients = NULL;
complex* _active_segment1_operator0_delta_a_field_coefficients = NULL;

// ********************************************************
//   field mg0_output globals
real* _mg0_output_t = NULL;
unsigned long _mg0_output_index_t = 0;

// vector raw globals
size_t _mg0_output_raw_alloc_size = 0;
real* _mg0_output_raw = NULL;
real* _active_mg0_output_raw = NULL;

// vector processed globals
size_t _mg0_output_processed_alloc_size = 0;
real* _mg0_output_processed = NULL;
real* _active_mg0_output_processed = NULL;
real* _mg0_output_sd = NULL;

// ********************************************************
//   field mg1_output globals
real* _mg1_output_t = NULL;
unsigned long _mg1_output_index_t = 0;

// vector raw globals
size_t _mg1_output_raw_alloc_size = 0;
real* _mg1_output_raw = NULL;
real* _active_mg1_output_raw = NULL;

// vector processed globals
size_t _mg1_output_processed_alloc_size = 0;
real* _mg1_output_processed = NULL;
real* _active_mg1_output_processed = NULL;
real* _mg1_output_sd = NULL;

// ********************************************************
//   field mg2_output globals
real* _mg2_output_t = NULL;
unsigned long _mg2_output_index_t = 0;

// vector raw globals
size_t _mg2_output_raw_alloc_size = 0;
real* _mg2_output_raw = NULL;
real* _active_mg2_output_raw = NULL;

// vector processed globals
size_t _mg2_output_processed_alloc_size = 0;
real* _mg2_output_processed = NULL;
real* _active_mg2_output_processed = NULL;
real* _mg2_output_sd = NULL;

// ********************************************************
//   field mg3_output globals
real* _mg3_output_t = NULL;
unsigned long _mg3_output_index_t = 0;

// vector raw globals
size_t _mg3_output_raw_alloc_size = 0;
real* _mg3_output_raw = NULL;
real* _active_mg3_output_raw = NULL;

// vector processed globals
size_t _mg3_output_processed_alloc_size = 0;
real* _mg3_output_processed = NULL;
real* _active_mg3_output_processed = NULL;
real* _mg3_output_sd = NULL;


// ********************************************************
// FUNCTION PROTOTYPES
// ********************************************************

// ********************************************************
//   field n function prototypes
void _n_coefficients_initialise();

// ********************************************************
//   field nx function prototypes
void _nx_oscillator_states_initialise();

// ********************************************************
//   field x function prototypes
void _x_densities_initialise();
void _x_densities_evaluate();

// ********************************************************
//   segment 0 (Top level sequence) function prototypes
void _segment0();

// ********************************************************
//   segment 1 (RK89 adaptive-step integrator) function prototypes
void _segment1();
inline void _segment1_calculate_delta_a(real _step);
inline void _segment1_ip_evolve(int _exponent);
inline void _segment1_calculate_nonconstant_ip_fields(real _step, int _exponent);
real _segment1_setup_sampling(bool* _next_sample_flag, long* _next_sample_counter);
real _segment1_n_coefficients_timestep_error(complex* _checkfield);
bool _segment1_n_coefficients_reset(complex* _reset_to);

void _segment1_n_operators_evaluate_operator0(real _step);
inline void _segment1_n_operators_operator0_copy_delta_a(real _step);

// ********************************************************
//   output function prototypes
void _write_output();

FILE* _open_xsil_file(const char* _filename);
void _close_xsil_file(FILE*& fp);
void _write_xsil_header(FILE* fp);
void _write_xsil_footer(FILE* fp);

// ********************************************************
//   moment group 0 function prototypes
void _mg0_sample();
void _mg0_process();
void _mg0_write_out(FILE* _outfile);

// ********************************************************
//   field mg0_output function prototypes
void _mg0_output_raw_initialise();

void _mg0_output_processed_initialise();

// ********************************************************
//   moment group 1 function prototypes
void _mg1_sample();
void _mg1_process();
void _mg1_write_out(FILE* _outfile);

// ********************************************************
//   field mg1_output function prototypes
void _mg1_output_raw_initialise();

void _mg1_output_processed_initialise();

// ********************************************************
//   moment group 2 function prototypes
void _mg2_sample();
void _mg2_process();
void _mg2_write_out(FILE* _outfile);

// ********************************************************
//   field mg2_output function prototypes
void _mg2_output_raw_initialise();

void _mg2_output_processed_initialise();

// ********************************************************
//   moment group 3 function prototypes
void _mg3_sample();
void _mg3_process();
void _mg3_write_out(FILE* _outfile);

// ********************************************************
//   field mg3_output function prototypes
void _mg3_output_raw_initialise();

void _mg3_output_processed_initialise();

// ********************************************************
// MAIN ROUTINE
// ********************************************************
int main(int argc, char **argv)
{
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &_rank);

  
  #ifdef __APPLE__
    #if __MAC_OS_X_VERSION_MIN_REQUIRED >= 1080
  {
    IOPMAssertionID _powerAssertionID = 0;
    IOReturn __io_result = IOPMAssertionCreateWithDescription(
      kIOPMAssertionTypePreventUserIdleSystemSleep,
      CFSTR("XMDS simulation 'Assignment1_program_solutionb_v1' preventing user idle sleep"), // Assertion name
      NULL, // Assertion details
      NULL, // Human-readable reason
      NULL, // Localization bundle path
      (CFTimeInterval)0, // never timeout
      kIOPMAssertionTimeoutActionRelease,
      &_powerAssertionID
      );
    if (__io_result != kIOReturnSuccess) {
      _LOG(_WARNING_LOG_LEVEL, "Failed to disable user idle sleep\n");
    }
    // Note, this power assertion is automatically released when the process quits.
  }
    #endif
  #endif
  
    
    
  
  _mg3_output_processed_alloc_size = MAX(_mg3_output_processed_alloc_size, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_processed_ncomponents);
  _mg3_output_raw_alloc_size = MAX(_mg3_output_raw_alloc_size, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_raw_ncomponents);
  _mg0_output_processed_alloc_size = MAX(_mg0_output_processed_alloc_size, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_processed_ncomponents);
  _nx_oscillator_states_alloc_size = MAX(_nx_oscillator_states_alloc_size, (_lattice_n * _lattice_x) * _nx_oscillator_states_ncomponents);
  _mg1_output_raw_alloc_size = MAX(_mg1_output_raw_alloc_size, (_mg1_output_lattice_t) * _mg1_output_raw_ncomponents);
  _mg2_output_processed_alloc_size = MAX(_mg2_output_processed_alloc_size, (_mg2_output_lattice_t) * _mg2_output_processed_ncomponents);
  _n_coefficients_alloc_size = MAX(_n_coefficients_alloc_size, (_lattice_n) * _n_coefficients_ncomponents);
  _mg0_output_raw_alloc_size = MAX(_mg0_output_raw_alloc_size, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_raw_ncomponents);
  _segment1_operator0_delta_a_field_coefficients_alloc_size = MAX(_segment1_operator0_delta_a_field_coefficients_alloc_size, (_lattice_n) * _segment1_operator0_delta_a_field_coefficients_ncomponents);
  _mg1_output_processed_alloc_size = MAX(_mg1_output_processed_alloc_size, (_mg1_output_lattice_t) * _mg1_output_processed_ncomponents);
  _mg2_output_raw_alloc_size = MAX(_mg2_output_raw_alloc_size, (_mg2_output_lattice_t) * _mg2_output_raw_ncomponents);
  _x_densities_alloc_size = MAX(_x_densities_alloc_size, (_lattice_x) * _x_densities_ncomponents);
  _n = (long*) xmds_malloc(sizeof(long) * (_lattice_n+1));
  
  _x = (real*) xmds_malloc(sizeof(real) * (_lattice_x+1));
  
  _kx = (real*) xmds_malloc(sizeof(real) * (_lattice_kx+1));
  
  _nprime = (long*) xmds_malloc(sizeof(long) * (_lattice_nprime+1));
  
  _n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _active_n_coefficients = _n_coefficients;
  
  _nx_oscillator_states = (real*) xmds_malloc(sizeof(real) * MAX(_nx_oscillator_states_alloc_size,1));
  _active_nx_oscillator_states = _nx_oscillator_states;
  
  _x_densities = (complex*) xmds_malloc(sizeof(complex) * MAX(_x_densities_alloc_size,1));
  _active_x_densities = _x_densities;
  _mg0_output_t = (real*) xmds_malloc(sizeof(real) * (_mg0_output_lattice_t+1));
  
  
  _mg0_output_raw = (real*) xmds_malloc(sizeof(real) * MAX(_mg0_output_raw_alloc_size,1));
  _active_mg0_output_raw = _mg0_output_raw;
  
  
  _mg0_output_processed = (real*) xmds_malloc(sizeof(real) * MAX(_mg0_output_processed_alloc_size,1));
  _active_mg0_output_processed = _mg0_output_processed;
  _mg0_output_sd = (real*) xmds_malloc(sizeof(real) * MAX(_mg0_output_processed_alloc_size,1)); // alias for _mg0_output_processed
  _mg1_output_t = (real*) xmds_malloc(sizeof(real) * (_mg1_output_lattice_t+1));
  
  
  _mg1_output_raw = (real*) xmds_malloc(sizeof(real) * MAX(_mg1_output_raw_alloc_size,1));
  _active_mg1_output_raw = _mg1_output_raw;
  
  
  _mg1_output_processed = (real*) xmds_malloc(sizeof(real) * MAX(_mg1_output_processed_alloc_size,1));
  _active_mg1_output_processed = _mg1_output_processed;
  _mg1_output_sd = (real*) xmds_malloc(sizeof(real) * MAX(_mg1_output_processed_alloc_size,1)); // alias for _mg1_output_processed
  _mg2_output_t = (real*) xmds_malloc(sizeof(real) * (_mg2_output_lattice_t+1));
  
  
  _mg2_output_raw = (real*) xmds_malloc(sizeof(real) * MAX(_mg2_output_raw_alloc_size,1));
  _active_mg2_output_raw = _mg2_output_raw;
  
  
  _mg2_output_processed = (real*) xmds_malloc(sizeof(real) * MAX(_mg2_output_processed_alloc_size,1));
  _active_mg2_output_processed = _mg2_output_processed;
  _mg2_output_sd = (real*) xmds_malloc(sizeof(real) * MAX(_mg2_output_processed_alloc_size,1)); // alias for _mg2_output_processed
  _mg3_output_t = (real*) xmds_malloc(sizeof(real) * (_mg3_output_lattice_t+1));
  
  
  _mg3_output_raw = (real*) xmds_malloc(sizeof(real) * MAX(_mg3_output_raw_alloc_size,1));
  _active_mg3_output_raw = _mg3_output_raw;
  
  
  _mg3_output_processed = (real*) xmds_malloc(sizeof(real) * MAX(_mg3_output_processed_alloc_size,1));
  _active_mg3_output_processed = _mg3_output_processed;
  _mg3_output_sd = (real*) xmds_malloc(sizeof(real) * MAX(_mg3_output_processed_alloc_size,1)); // alias for _mg3_output_processed
  
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++)
    _n[_index_n] = _min_n + _index_n*_dn;
  for (long _index_x = 0; _index_x < _lattice_x; _index_x++)
    _x[_index_x] = _min_x + _index_x*_dx;
  for (long _index_kx = 0; _index_kx < (_lattice_kx+1)/2; _index_kx++)
    _kx[_index_kx] = _index_kx*_dkx;
  for (long _index_kx = (_lattice_kx+1)/2; _index_kx < _lattice_kx; _index_kx++)
    _kx[_index_kx] = -(_lattice_kx - _index_kx) * _dkx;
  for (long _index_nprime = 0; _index_nprime < _lattice_nprime; _index_nprime++)
    _nprime[_index_nprime] = _min_nprime + _index_nprime*_dnprime;
  _active_mg0_output_raw = _mg0_output_raw;
  _mg0_output_raw_initialise();
  
  _active_mg0_output_processed = _mg0_output_sd;
  _mg0_output_processed_initialise();
  _active_mg0_output_processed = _mg0_output_processed;
  _mg0_output_processed_initialise();
  _active_mg1_output_raw = _mg1_output_raw;
  _mg1_output_raw_initialise();
  
  _active_mg1_output_processed = _mg1_output_sd;
  _mg1_output_processed_initialise();
  _active_mg1_output_processed = _mg1_output_processed;
  _mg1_output_processed_initialise();
  _active_mg2_output_raw = _mg2_output_raw;
  _mg2_output_raw_initialise();
  
  _active_mg2_output_processed = _mg2_output_sd;
  _mg2_output_processed_initialise();
  _active_mg2_output_processed = _mg2_output_processed;
  _mg2_output_processed_initialise();
  _active_mg3_output_raw = _mg3_output_raw;
  _mg3_output_raw_initialise();
  
  _active_mg3_output_processed = _mg3_output_sd;
  _mg3_output_processed_initialise();
  _active_mg3_output_processed = _mg3_output_processed;
  _mg3_output_processed_initialise();
  // load wisdom
  #if CFG_OSAPI == CFG_OSAPI_POSIX // Don't load wisdom on windows
  {
    char _hostName[256];
    gethostname(_hostName, 256);
    _hostName[255] = '\0'; // just in case
    
    string _pathToWisdom = getenv("HOME");
    _pathToWisdom += "/.xmds/wisdom/";
    
    string _wisdomFileName = _hostName;
    _wisdomFileName += ".wisdom";
    _wisdomFileName += ".fftw3";
    
    FILE *_fp = NULL;
    
    _fp = fopen(_pathToWisdom.c_str(), "r");
    if (_fp) {
      fclose(_fp);
    } else {
      int _result = mkdir((string(getenv("HOME")) + "/.xmds").c_str(), S_IRWXU);
      if (mkdir(_pathToWisdom.c_str(), S_IRWXU)) {
        // We failed to create the ~/.xmds/wisdom directory
        _LOG(_WARNING_LOG_LEVEL, "Warning: Cannot find enlightenment, the path to wisdom ~/.xmds/wisdom doesn't seem to exist and we couldn't create it.\n"
                                 "         I'll use the current path instead.\n");
        _pathToWisdom = ""; // present directory
      }
      
    }
    
    _fftwWisdomPath = _pathToWisdom + _wisdomFileName;
    
    FILE *_wisdomFile = NULL;
    if ( (_wisdomFile = fopen(_fftwWisdomPath.c_str(), "r")) != NULL) {
      _LOG(_SIMULATION_LOG_LEVEL, "Found enlightenment... (Importing wisdom)\n");
      fftw_import_wisdom_from_file(_wisdomFile);
      fclose(_wisdomFile);
    }
  }
  #endif // POSIX
  
  _basis_transform_t *_basis_transform = NULL;
  ptrdiff_t _auxiliary_array_size = 0;
  ptrdiff_t _max_vector_size = 0;
  real* _max_vector_array = NULL;
  
  if (_auxiliary_array_size) {
    _auxiliary_array = (real*) xmds_malloc(sizeof(real) * _auxiliary_array_size);
  }
  
  bool _allocated_temporary_array = false;
  if (!_max_vector_array && _max_vector_size > 0) {
    _max_vector_array = (real*) xmds_malloc(sizeof(real) * _max_vector_size);
    _allocated_temporary_array = true;
  }
  
  // Make all geometry-dependent transformations prepare plans, etc.
  
  if (_allocated_temporary_array) {
    xmds_free(_max_vector_array);
  }
  
  // Get the time at which the simulation started
  timeval _tim;
  gettimeofday(&_tim, NULL);
  double _startTime = _tim.tv_sec + (_tim.tv_usec/1e6);
  
  /* Code that actually does stuff goes here */
  _segment0();
  
  
  _write_output();
  
  // Work out how long the simulation has run for
  gettimeofday(&_tim, NULL);
  double _endTime = _tim.tv_sec + (_tim.tv_usec/1e6);
  _LOG(_SIMULATION_LOG_LEVEL, "Time elapsed for simulation is: %.2f seconds\n", _endTime - _startTime);
  if (_auxiliary_array) {
    xmds_free(_auxiliary_array);
  }
  
  // Save wisdom
  #if CFG_OSAPI == CFG_OSAPI_POSIX
  {
    FILE *_wisdomFile = NULL;
    if ( (_wisdomFile = fopen(_fftwWisdomPath.c_str(), "w")) != NULL) {
      fftw_export_wisdom_to_file(_wisdomFile);
      fclose(_wisdomFile);
    }
  }
  #endif // POSIX
  
  fftw_cleanup();
  
  xmds_free(_n_coefficients);
  _active_n_coefficients = _n_coefficients = NULL;
  
  xmds_free(_nx_oscillator_states);
  _active_nx_oscillator_states = _nx_oscillator_states = NULL;
  
  xmds_free(_x_densities);
  _active_x_densities = _x_densities = NULL;
  
  xmds_free(_mg0_output_raw);
  _active_mg0_output_raw = _mg0_output_raw = NULL;
  
  
  xmds_free(_mg0_output_processed);
  _active_mg0_output_processed = _mg0_output_processed = NULL;
  xmds_free(_mg0_output_sd); // alias for _mg0_output_processed
  _mg0_output_sd = NULL;
  
  xmds_free(_mg1_output_raw);
  _active_mg1_output_raw = _mg1_output_raw = NULL;
  
  
  xmds_free(_mg1_output_processed);
  _active_mg1_output_processed = _mg1_output_processed = NULL;
  xmds_free(_mg1_output_sd); // alias for _mg1_output_processed
  _mg1_output_sd = NULL;
  
  xmds_free(_mg2_output_raw);
  _active_mg2_output_raw = _mg2_output_raw = NULL;
  
  
  xmds_free(_mg2_output_processed);
  _active_mg2_output_processed = _mg2_output_processed = NULL;
  xmds_free(_mg2_output_sd); // alias for _mg2_output_processed
  _mg2_output_sd = NULL;
  
  xmds_free(_mg3_output_raw);
  _active_mg3_output_raw = _mg3_output_raw = NULL;
  
  
  xmds_free(_mg3_output_processed);
  _active_mg3_output_processed = _mg3_output_processed = NULL;
  xmds_free(_mg3_output_sd); // alias for _mg3_output_processed
  _mg3_output_sd = NULL;
  
  
  MPI_Finalize();
  
  return 0;
}

// ********************************************************
// FUNCTION IMPLEMENTATIONS
// ********************************************************

inline void *xmds_malloc(size_t size)
{
  void *retPointer = _xmds_malloc(size);
  if ( !retPointer )
    _LOG(_ERROR_LOG_LEVEL, "ERROR: Couldn't allocate %zu bytes of memory!", size);
  return retPointer;
}


// ********************************************************
//   MPI Multipath Simulation Driver function implementations
void _segment0()
{
  
  
  _segment1_operator0_delta_a_field_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_segment1_operator0_delta_a_field_coefficients_alloc_size,1));
  _active_segment1_operator0_delta_a_field_coefficients = _segment1_operator0_delta_a_field_coefficients;
  _segment1_akafield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akbfield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akcfield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akdfield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akefield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akffield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akgfield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akhfield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akifield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_akjfield_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  _segment1_initial_n_coefficients = (complex*) xmds_malloc(sizeof(complex) * MAX(_n_coefficients_alloc_size,1));
  
  
  for (long _i0 = _rank; _i0 < _n_paths; _i0+=_size) {
    _LOG(_PATH_LOG_LEVEL, "Starting path %li\n", _i0 + 1);
  
  
  
  
    _mg0_output_raw_initialise();  _mg1_output_raw_initialise();  _mg2_output_raw_initialise();  _mg3_output_raw_initialise();
    t = 0.0;
    
    _mg0_output_raw_initialise();
    _mg1_output_raw_initialise();
    _mg2_output_raw_initialise();
    _mg3_output_raw_initialise();
    _active_n_coefficients = _n_coefficients;
    _n_coefficients_initialise();
    _active_nx_oscillator_states = _nx_oscillator_states;
    _nx_oscillator_states_initialise();
    _mg0_output_index_t = 0;
    _mg1_output_index_t = 0;
    _mg2_output_index_t = 0;
    _mg3_output_index_t = 0;
    _mg0_sample();
    _mg1_sample();
    _mg2_sample();
    _mg3_sample();
    _segment1();
    
    _mg0_process();
    _mg1_process();
    _mg2_process();
    _mg3_process();
  }
  
  
  
  xmds_free(_segment1_operator0_delta_a_field_coefficients);
  _active_segment1_operator0_delta_a_field_coefficients = _segment1_operator0_delta_a_field_coefficients = NULL;
  xmds_free(_segment1_akafield_n_coefficients);
  xmds_free(_segment1_akbfield_n_coefficients);
  xmds_free(_segment1_akcfield_n_coefficients);
  xmds_free(_segment1_akdfield_n_coefficients);
  xmds_free(_segment1_akefield_n_coefficients);
  xmds_free(_segment1_akffield_n_coefficients);
  xmds_free(_segment1_akgfield_n_coefficients);
  xmds_free(_segment1_akhfield_n_coefficients);
  xmds_free(_segment1_akifield_n_coefficients);
  xmds_free(_segment1_akjfield_n_coefficients);
  xmds_free(_segment1_initial_n_coefficients);
  
  
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg0_output_raw, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg0_output_raw, NULL, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg0_output_processed, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg0_output_processed, NULL, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg0_output_sd, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg0_output_sd, NULL, (_mg0_output_lattice_t * _lattice_n) * _mg0_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg1_output_raw, (_mg1_output_lattice_t) * _mg1_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg1_output_raw, NULL, (_mg1_output_lattice_t) * _mg1_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg1_output_processed, (_mg1_output_lattice_t) * _mg1_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg1_output_processed, NULL, (_mg1_output_lattice_t) * _mg1_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg1_output_sd, (_mg1_output_lattice_t) * _mg1_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg1_output_sd, NULL, (_mg1_output_lattice_t) * _mg1_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg2_output_raw, (_mg2_output_lattice_t) * _mg2_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg2_output_raw, NULL, (_mg2_output_lattice_t) * _mg2_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg2_output_processed, (_mg2_output_lattice_t) * _mg2_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg2_output_processed, NULL, (_mg2_output_lattice_t) * _mg2_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg2_output_sd, (_mg2_output_lattice_t) * _mg2_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg2_output_sd, NULL, (_mg2_output_lattice_t) * _mg2_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg3_output_raw, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg3_output_raw, NULL, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_raw_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg3_output_processed, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg3_output_processed, NULL, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    
  if (_rank == 0)
    MPI_Reduce(MPI_IN_PLACE, _mg3_output_sd, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  else
    MPI_Reduce(_mg3_output_sd, NULL, (_mg3_output_lattice_t * _lattice_x) * _mg3_output_processed_ncomponents,
               MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
}


// ********************************************************
//   field n function implementations
// initialisation for vector coefficients
void _n_coefficients_initialise()
{
  
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    // The purpose of the following define is to give a (somewhat helpful) compile-time error
    // if the user has attempted to use the propagation dimension variable in the initialisation
    // block of a <vector> element. If they're trying to do this, what they really want is a 
    // <computed_vector> instead.
    #define t Dont_use_propagation_dimension_t_in_vector_element_CDATA_block___Use_a_computed_vector_instead
    
    // ********** Initialisation code ***************
    #line 85 "Assignment1_program_solutionb_v1.xmds"
    
    cup = 0.0;
    cdown = 0.0;
    
    if(n==0){//oscillator in the ground-state
       cup = c1;
       cdown = c2;
    }    
    
    #line 1146 "Assignment1_program_solutionb_v1.cc"
    // **********************************************
    #undef t
    
    // Increment index pointers for vectors in field n (or having the same dimensions)
    _n_coefficients_index_pointer += 1 * _n_coefficients_ncomponents;
    
  }
  #undef n
  #undef dn
  #undef cup
  #undef cdown
}

// ********************************************************
//   field nx function implementations
// initialisation for vector oscillator_states
void _nx_oscillator_states_initialise()
{
  
  long _nx_oscillator_states_index_pointer = 0;
  #define varphi _active_nx_oscillator_states[_nx_oscillator_states_index_pointer + 0]
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    #define x _x[_index_x + 0]
    #define dx (_dx * (1.0))
    
    for (long _index_x = 0; _index_x < _lattice_x; _index_x++) {
      // The purpose of the following define is to give a (somewhat helpful) compile-time error
      // if the user has attempted to use the propagation dimension variable in the initialisation
      // block of a <vector> element. If they're trying to do this, what they really want is a 
      // <computed_vector> instead.
      #define t Dont_use_propagation_dimension_t_in_vector_element_CDATA_block___Use_a_computed_vector_instead
      
      // ********** Initialisation code ***************
      #line 100 "Assignment1_program_solutionb_v1.xmds"
      
      double normfact = 1.0/sqrt(sqrt(M_PI)*sigmaosc*factorial(n)*integerpower(2,n));
      varphi = normfact*exp(-0.5*x*x/sigmaosc/sigmaosc)*Hermite(x/sigmaosc,n);
      
      #line 1188 "Assignment1_program_solutionb_v1.cc"
      // **********************************************
      #undef t
      
      // Increment index pointers for vectors in field nx (or having the same dimensions)
      _nx_oscillator_states_index_pointer += 1 * _nx_oscillator_states_ncomponents;
      
    }
    #undef x
    #undef dx
  }
  #undef n
  #undef dn
  #undef varphi
}

// ********************************************************
//   field x function implementations
// initialisation for computed vector densities
void _x_densities_initialise()
{
  // Because we're integrating over dimensions, we need to set the vector to zero.
  bzero(_active_x_densities, sizeof(complex) * _x_densities_alloc_size);
}


void _x_densities_evaluate()
{
  _active_x_densities = _x_densities;
  _x_densities_initialise();
  
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  long _nx_oscillator_states_index_pointer = 0;
  #define varphi _active_nx_oscillator_states[_nx_oscillator_states_index_pointer + 0]
  long _x_densities_index_pointer = 0;
  complex phiposup;
  complex phiposdown;
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    #define x _x[_index_x + 0]
    #define dx (_dx * (1.0))
    
    for (long _index_x = 0; _index_x < _lattice_x; _index_x++) {
      // Set index pointers explicitly for (some) vectors
      _n_coefficients_index_pointer = ( 0
         + _index_n * 1 ) * _n_coefficients_ncomponents;
      _x_densities_index_pointer = ( 0
         + _index_x * 1 ) * _x_densities_ncomponents;
      // ************* Evaluation code ****************
      #line 111 "Assignment1_program_solutionb_v1.xmds"
      
      phiposup = varphi*cup;
      phiposdown = varphi*cdown;
      
      #line 1246 "Assignment1_program_solutionb_v1.cc"
      // **********************************************
      
      _active_x_densities[_x_densities_index_pointer + 0] += phiposup * dn;
      _active_x_densities[_x_densities_index_pointer + 1] += phiposdown * dn;
      // Increment index pointers for vectors in field nx (or having the same dimensions)
      _nx_oscillator_states_index_pointer += 1 * _nx_oscillator_states_ncomponents;
      
    }
    #undef x
    #undef dx
  }
  #undef n
  #undef dn
  #undef cup
  #undef cdown
  #undef varphi
}

// ********************************************************
//   segment 1 (RK89 adaptive-step integrator) function implementations
void _segment1()
{
  real _step = 20/(real)1000;
  real _old_step = _step;
  real _min_step = _step;
  real _max_step = _step;
  long _attempted_steps = 0;
  long _unsuccessful_steps = 0;
  
  real _tolerance = 1e-08;
  
  real _error, _last_norm_error = 1.0;
  real _segment1_n_coefficients_error;
  
  bool _discard = false;
  bool _break_next = false;
  
  bool _next_sample_flag[6];
  for (long _i0 = 0; _i0 < 6; _i0++)
    _next_sample_flag[_i0] = false;
  
  long _next_sample_counter[4];
  for (long _i0 = 0; _i0 < 4; _i0++)
    _next_sample_counter[_i0] = 1;
  
  real _t_local = 0.0;
  
  real _t_break_next = _segment1_setup_sampling(_next_sample_flag, _next_sample_counter);
  
  if ( (_t_local + _step)*(1.0 + _EPSILON) >= _t_break_next) {
    _break_next = true;
    _step = _t_break_next - _t_local;
  }
  
  complex* _akafield_n_coefficients = _segment1_akafield_n_coefficients;
  complex* _akbfield_n_coefficients = _segment1_akbfield_n_coefficients;
  complex* _akcfield_n_coefficients = _segment1_akcfield_n_coefficients;
  complex* _akdfield_n_coefficients = _segment1_akdfield_n_coefficients;
  complex* _akefield_n_coefficients = _segment1_akefield_n_coefficients;
  complex* _akffield_n_coefficients = _segment1_akffield_n_coefficients;
  complex* _akgfield_n_coefficients = _segment1_akgfield_n_coefficients;
  complex* _akhfield_n_coefficients = _segment1_akhfield_n_coefficients;
  complex* _akifield_n_coefficients = _segment1_akifield_n_coefficients;
  complex* _akjfield_n_coefficients = _segment1_akjfield_n_coefficients;
  complex* _initial_n_coefficients = _segment1_initial_n_coefficients;
  
  
  // Runge Kutta method constants 
  real _a_raw[16];
  real _a[16];
  real _b[16][16];
  real _c[16];
  real _cs[16];
  real _d[16];
  
  for (unsigned long _i0 = 0; _i0 < 16; _i0++) {
    _a_raw[_i0] = _c[_i0] = _d[_i0] = 0.0;
    for (unsigned long _i1 = 0; _i1 < 16; _i1++)
     _b[_i0][_i1] = 0.0;
  }
  
  _a_raw[1]  = 0.02173913043478260869565217391304347;
  _a_raw[2]  = 0.09629581047800066670113001679819925;
  _a_raw[3]  = 0.14444371571700100005169502519729888;
  _a_raw[4]  = 0.52205882352941176470588235294117647;
  _a_raw[5]  = 0.22842443612863469578031459099794265;
  _a_raw[6]  = 0.54360353589933733219171338103002937;
  _a_raw[7]  = 0.64335664335664335664335664335664335;
  _a_raw[8]  = 0.48251748251748251748251748251748251;
  _a_raw[9]  = 0.06818181818181818181818181818181818;
  _a_raw[10] = 0.25060827250608272506082725060827250;
  _a_raw[11] = 0.66736715965600568968278165443304378;
  _a_raw[12] = 0.85507246376811594202898550724637681;
  _a_raw[13] = 0.89795918367346938775510204081632653;
  _a_raw[14] = 1.0;
  _a_raw[15] = 1.0;
  
  _a[0] = 0.0;
  for (unsigned long _i0 = 1; _i0 < 16; _i0++)
    _a[_i0] = _a_raw[_i0] - _a_raw[_i0 - 1];
  
  _b[1][0]   = 1.0/46.0;
  _b[2][0]   =-0.11698050118114486205818241524969622;
  _b[2][1]   = 0.21327631165914552875931243204789548;
  _b[3][0]   = 0.03611092892925025001292375629932472;
  _b[3][2]   = 0.10833278678775075003877126889797416;
  _b[4][0]   = 1.57329743908138605107331820072051125;
  _b[4][2]   =-5.98400943754042002888532938159655553;
  _b[4][3]   = 4.93277082198844574251789353381722074;
  _b[5][0]   = 0.05052046351120380909008334360006234;
  _b[5][3]   = 0.17686653884807108146683657390397612;
  _b[5][4]   = 0.00103743376935980522339467349390418;
  _b[6][0]   = 0.10543148021953768958529340893598138;
  _b[6][3]   =-0.16042415162569842979496486916719383;
  _b[6][4]   = 0.11643956912829316045688724281285250;
  _b[6][5]   = 0.48215663817720491194449759844838932;
  _b[7][0]   = 0.07148407148407148407148407148407148;
  _b[7][5]   = 0.32971116090443908023196389566296464;
  _b[7][6]   = 0.24216141096813279233990867620960722;
  _b[8][0]   = 0.07162368881118881118881118881118881;
  _b[8][5]   = 0.32859867301674234161492268975519694;
  _b[8][6]   = 0.11622213117906185418927311444060725;
  _b[8][7]   =-0.03392701048951048951048951048951048;
  _b[9][0]   = 0.04861540768024729180628870095388582;
  _b[9][5]   = 0.03998502200331629058445317782406268;
  _b[9][6]   = 0.10715724786209388876739304914053506;
  _b[9][7]   =-0.02177735985419485163815426357369818;
  _b[9][8]   =-0.10579849950964443770179884616296721;
  _b[10][0]  =-0.02540141041535143673515871979014924;
  _b[10][5]  = 1.0/30.0;
  _b[10][6]  =-0.16404854760069182073503553020238782;
  _b[10][7]  = 0.03410548898794737788891414566528526;
  _b[10][8]  = 0.15836825014108792658008718465091487;
  _b[10][9]  = 0.21425115805975734472868683695127609;
  _b[11][0]  = 0.00584833331460742801095934302256470;
  _b[11][5]  =-0.53954170547283522916525526480339109;
  _b[11][6]  = 0.20128430845560909506500331018201158;
  _b[11][7]  = 0.04347222773254789483240207937678906;
  _b[11][8]  =-0.00402998571475307250775349983910179;
  _b[11][9]  = 0.16541535721570612771420482097898952;
  _b[11][10] = 0.79491862412512344573322086551518180;
  _b[12][0]  =-0.39964965968794892497157706711861448;
  _b[12][5]  =-3.79096577568393158554742638116249372;
  _b[12][6]  =-0.40349325653530103387515807815498044;
  _b[12][7]  =-2.82463879530435263378049668286220715;
  _b[12][8]  = 1.04226892772185985533374283289821416;
  _b[12][9]  = 1.12510956420436603974237036536924078;
  _b[12][10] = 3.32746188718986816186934832571938138;
  _b[12][11] = 2.77897957186355606325818219255783627;
  _b[13][0]  = 0.39545306350085237157098218205756922;
  _b[13][5]  = 5.82534730759650564865380791881446903;
  _b[13][6]  =-0.36527452339161313311889856846974452;
  _b[13][7]  = 1.18860324058346533283780076203192232;
  _b[13][8]  = 0.57970467638357921347110271762687972;
  _b[13][9]  =-0.86824862589087693262676988867897834;
  _b[13][10] =-5.20227677296454721392873650976792184;
  _b[13][11] =-0.79895541420753382543211121058675915;
  _b[13][12] = 0.14360623206363792632792463778889008;
  _b[14][0]  = 8.49173149061346398013352206978380938;
  _b[14][5]  = 86.32213734729036800877634194386790750;
  _b[14][6]  = 1.02560575501091662034511526187393241;
  _b[14][7]  = 85.77427969817339941806831550695235092;
  _b[14][8]  =-13.98699305104110611795532466113248067;
  _b[14][9]  =-20.71537405501426352265946477613161883;
  _b[14][10] =-72.16597156619946800281180102605140463;
  _b[14][11] =-76.71211139107806345587696023064419687;
  _b[14][12] = 4.22319427707298828839851258893735507;
  _b[14][13] =-1.25649850482823521641825667745565428;
  _b[15][0]  =-0.42892119881959353241190195318730008;
  _b[15][5]  =-9.16865700950084689999297912545025359;
  _b[15][6]  = 1.08317616770620939241547721530003920;
  _b[15][7]  =-1.23501525358323653198215832293981810;
  _b[15][8]  =-1.21438272617593906232943856422371019;
  _b[15][9]  = 1.37226168507232166621351243731869914;
  _b[15][10] = 9.15723239697162418155377135344394113;
  _b[15][11] = 1.30616301842220047563298585480401671;
  _b[15][12] =-0.25285618808937955976690569433069974;
  _b[15][13] = 0.38099910799663987066763679926508552;
  
  _c[0]  = 0.01490902081978461022483617102382552;
  _c[7]  =-0.20408044692054151258349120934134791;
  _c[8]  = 0.22901438600570447264772469337066476;
  _c[9]  = 0.12800558251147375669208211573729202;
  _c[10] = 0.22380626846054143649770066956485937;
  _c[11] = 0.39553165293700054420552389156421651;
  _c[12] = 0.05416646758806981196568364538360743;
  _c[13] = 0.12691439652445903685643385312168037;
  _c[14] =-0.00052539244262118876455834655383035;
  _c[15] = 1.0/31.0;
  
  _cs[0]  = 0.00653047880643482012034413441159249;
  _cs[7]  =-2.31471038197461347517552506241529830;
  _cs[8]  = 0.43528227238866280799530900822377013;
  _cs[9]  = 0.14907947287101933118545845390618763;
  _cs[10] = 0.17905535442235532311850533252768020;
  _cs[11] = 2.53400872222767706921176214508820825;
  _cs[12] =-0.55430437423209112896721332268159015;
  _cs[13] = 0.56924788787870083224213506297615260;
  _cs[14] =-0.03644749690427461198884026816573513;
  _cs[15] = 1.0/31.0;
  
  _d[0] = 1.0-_b[15][5]/_b[14][5];
  _d[1] = _b[15][0]-_b[14][0]*_b[15][5]/_b[14][5];
  _d[2] = _b[15][5]/_b[14][5];
  _d[3] = _b[15][6]-_b[14][6]*_b[15][5]/_b[14][5];
  _d[4] = _b[15][7]-_b[14][7]*_b[15][5]/_b[14][5];
  _d[5] = _b[15][8]-_b[14][8]*_b[15][5]/_b[14][5];
  _d[6] = _b[15][9]-_b[14][9]*_b[15][5]/_b[14][5];
  _d[7] = _b[15][10]-_b[14][10]*_b[15][5]/_b[14][5];
  _d[8] = _b[15][11]-_b[14][11]*_b[15][5]/_b[14][5];
  _d[9] = _b[15][12]-_b[14][12]*_b[15][5]/_b[14][5];
  _d[10] = _b[15][13]-_b[14][13]*_b[15][5]/_b[14][5];
  
  do {
    
    do {
      
      
      // Step 1
      
      memcpy(_akafield_n_coefficients, _n_coefficients, sizeof(complex) * _n_coefficients_alloc_size);
      
      _segment1_calculate_nonconstant_ip_fields(_step, 1);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(1);
      
      _active_n_coefficients = _akafield_n_coefficients;
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(1);
      
      // Step 2
      
      t += _a[1] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akbfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[1][0]*_AUTOVEC(_akafield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akbfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 2);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-2);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(2);
      
      // Step 3
      
      t += _a[2] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akcfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[2][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[2][1]*_AUTOVEC(_akbfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akcfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 3);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-3);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(3);
      
      // Step 4
      
      t += _a[3] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akdfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[3][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[3][1]*_AUTOVEC(_akbfield_n_coefficients)[_i0]
              + _b[3][2]*_AUTOVEC(_akcfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akdfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 4);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-4);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(4);
      
      // Step 5
      
      t += _a[4] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akefield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[4][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[4][1]*_AUTOVEC(_akbfield_n_coefficients)[_i0]
              + _b[4][2]*_AUTOVEC(_akcfield_n_coefficients)[_i0] + _b[4][3]*_AUTOVEC(_akdfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akefield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 5);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-5);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(5);
      
      // Step 6
      
      t += _a[5] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akifield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[5][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[5][3]*_AUTOVEC(_akdfield_n_coefficients)[_i0]
              + _b[5][4]*_AUTOVEC(_akefield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akifield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 6);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-6);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(6);
      
      // Step 7
      
      t += _a[6] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akjfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[6][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[6][3]*_AUTOVEC(_akdfield_n_coefficients)[_i0]
              + _b[6][4]*_AUTOVEC(_akefield_n_coefficients)[_i0] + _b[6][5]*_AUTOVEC(_akifield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akjfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 7);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-7);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(7);
      
      // Step 8
      
      t += _a[7] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akbfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[7][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[7][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[7][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akbfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 8);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-8);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(8);
      
      // Step 9
      
      t += _a[8] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akcfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[8][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[8][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[8][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0]+ _b[8][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akcfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 9);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-9);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(9);
      
      // Step 10
      
      t += _a[9] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akdfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[9][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[9][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[9][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0]+ _b[9][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0]+ _b[9][8]*_AUTOVEC(_akcfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akdfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 10);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-10);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(10);
      
      // Step 11
      
      t += _a[10] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akefield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[10][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[10][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[10][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0]+ _b[10][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0] + _b[10][8]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
              + _b[10][9]*_AUTOVEC(_akdfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akefield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 11);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-11);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(11);
      
      // Step 12
      
      t += _a[11] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akffield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[11][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[11][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[11][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0] + _b[11][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0] + _b[11][8]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
              + _b[11][9]*_AUTOVEC(_akdfield_n_coefficients)[_i0] + _b[11][10]*_AUTOVEC(_akefield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akffield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 12);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-12);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(12);
      
      // Step 13
      
      t += _a[12] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akgfield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akgfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[12][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[12][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[12][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0]+ _b[12][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0] + _b[12][8]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
              + _b[12][9]*_AUTOVEC(_akdfield_n_coefficients)[_i0] + _b[12][10]*_AUTOVEC(_akefield_n_coefficients)[_i0] + _b[12][11]*_AUTOVEC(_akffield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akgfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 13);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-13);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(13);
      
      // Step 14
      
      t += _a[13] * _step;
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akhfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akgfield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akhfield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[13][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[13][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[13][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0]+ _b[13][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0] + _b[13][8]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
              + _b[13][9]*_AUTOVEC(_akdfield_n_coefficients)[_i0] + _b[13][10]*_AUTOVEC(_akefield_n_coefficients)[_i0] + _b[13][11]*_AUTOVEC(_akffield_n_coefficients)[_i0]
              + _b[13][12]*_AUTOVEC(_akgfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _akhfield_n_coefficients;
      
      _segment1_calculate_nonconstant_ip_fields(_step, 14);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(-14);
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // a_i=D(a_2*dt)[y1]
      _segment1_ip_evolve(14);
      
      // Step 15 and 16 combined to reduce memory use 
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akhfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akgfield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akifield_n_coefficients)[_i0] = _AUTOVEC(_n_coefficients)[_i0] + _b[14][0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _b[14][5]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _b[14][6]*_AUTOVEC(_akjfield_n_coefficients)[_i0]+ _b[14][7]*_AUTOVEC(_akbfield_n_coefficients)[_i0] + _b[14][8]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
              + _b[14][9]*_AUTOVEC(_akdfield_n_coefficients)[_i0] + _b[14][10]*_AUTOVEC(_akefield_n_coefficients)[_i0] + _b[14][11]*_AUTOVEC(_akffield_n_coefficients)[_i0]
              + _b[14][12]*_AUTOVEC(_akgfield_n_coefficients)[_i0] + _b[14][13]*_AUTOVEC(_akhfield_n_coefficients)[_i0];
        }
      
      }
      
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akhfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akgfield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akjfield_n_coefficients)[_i0] = _d[0]*_AUTOVEC(_n_coefficients)[_i0]
                + _d[1]*_AUTOVEC(_akafield_n_coefficients)[_i0]
                + _d[2]*_AUTOVEC(_akifield_n_coefficients)[_i0]
                + _d[3]*_AUTOVEC(_akjfield_n_coefficients)[_i0]
                + _d[4]*_AUTOVEC(_akbfield_n_coefficients)[_i0]
                + _d[5]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
                + _d[6]*_AUTOVEC(_akdfield_n_coefficients)[_i0]
                + _d[7]*_AUTOVEC(_akefield_n_coefficients)[_i0]
                + _d[8]*_AUTOVEC(_akffield_n_coefficients)[_i0]
                + _d[9]*_AUTOVEC(_akgfield_n_coefficients)[_i0]
                + _d[10]*_AUTOVEC(_akhfield_n_coefficients)[_i0];
        }
      
      }
      
      t += _a[14] * _step;
      
      _active_n_coefficients = _akifield_n_coefficients;
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      t += _a[15] * _step;
      
      _active_n_coefficients = _akjfield_n_coefficients;
      
      // a_k=G[a_k, t]
      _segment1_calculate_delta_a(_step);
      
      // Take full step
      
      // ai = a
      memcpy(_initial_n_coefficients, _n_coefficients, sizeof(complex) * _n_coefficients_alloc_size);
      
      // a = a + etc
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akhfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akgfield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_n_coefficients)[_i0] += _c[0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _c[7]*_AUTOVEC(_akbfield_n_coefficients)[_i0] + _c[8]*_AUTOVEC(_akcfield_n_coefficients)[_i0]
              + _c[9]*_AUTOVEC(_akdfield_n_coefficients)[_i0] + _c[10]*_AUTOVEC(_akefield_n_coefficients)[_i0] + _c[11]*_AUTOVEC(_akffield_n_coefficients)[_i0]
              + _c[12]*_AUTOVEC(_akgfield_n_coefficients)[_i0] + _c[13]*_AUTOVEC(_akhfield_n_coefficients)[_i0] + _c[14]*_AUTOVEC(_akifield_n_coefficients)[_i0]
              + _c[15]*_AUTOVEC(_akjfield_n_coefficients)[_i0];
        }
      
      }
      
      // a* = a + etc
      {
        _MAKE_AUTOVEC_VARIABLE(_akdfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akffield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akhfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akjfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_initial_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akcfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akbfield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akefield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akafield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akifield_n_coefficients);
        _MAKE_AUTOVEC_VARIABLE(_akgfield_n_coefficients);
        #pragma ivdep
        for (long _i0 = 0; _i0 < 2 * (_lattice_n) * _n_coefficients_ncomponents; _i0++) {
          _AUTOVEC(_akafield_n_coefficients)[_i0] = _AUTOVEC(_initial_n_coefficients)[_i0] + _cs[0]*_AUTOVEC(_akafield_n_coefficients)[_i0] + _cs[7]*_AUTOVEC(_akbfield_n_coefficients)[_i0]
              + _cs[8]*_AUTOVEC(_akcfield_n_coefficients)[_i0] + _cs[9]*_AUTOVEC(_akdfield_n_coefficients)[_i0] + _cs[10]*_AUTOVEC(_akefield_n_coefficients)[_i0]
              + _cs[11]*_AUTOVEC(_akffield_n_coefficients)[_i0] + _cs[12]*_AUTOVEC(_akgfield_n_coefficients)[_i0] + _cs[13]*_AUTOVEC(_akhfield_n_coefficients)[_i0]
              + _cs[14]*_AUTOVEC(_akifield_n_coefficients)[_i0] + _cs[15]*_AUTOVEC(_akjfield_n_coefficients)[_i0];
        }
      
      }
      
      _active_n_coefficients = _n_coefficients;
      
      
      
      _error = 0.0;
      
      _segment1_n_coefficients_error = _segment1_n_coefficients_timestep_error(_akafield_n_coefficients);
      if (_segment1_n_coefficients_error > _error)
        _error = _segment1_n_coefficients_error;
      
      _attempted_steps++;
      
      if (_error < _tolerance) {
        _t_local += _step;
        if (_step > _max_step)
          _max_step = _step;
        if (!_break_next && _step < _min_step)
          _min_step = _step;
        _discard = false;
      } else {
        t -= _step;
  
        if (_segment1_n_coefficients_reset(_initial_n_coefficients) == false) {
  
          _LOG(_WARNING_LOG_LEVEL, "WARNING: NaN present. Integration halted at t = %e.\n"
                             "         Non-finite number in integration vector \"coefficients\" in segment 1.\n", t);
          if (_mg0_output_index_t < _mg0_output_lattice_t)
            _mg0_sample();
          if (_mg1_output_index_t < _mg1_output_lattice_t)
            _mg1_sample();
          if (_mg2_output_index_t < _mg2_output_lattice_t)
            _mg2_sample();
          if (_mg3_output_index_t < _mg3_output_lattice_t)
            _mg3_sample();
          
          goto _SEGMENT1_END;
        }
  
        _segment1_ip_evolve(-1);
        
        _discard = true;
        _break_next = false;
        _unsuccessful_steps++;
      }
      
      _old_step = _step;
      
      // Resize step
      if (_error < 0.5*_tolerance || _error > _tolerance) {
        const real _safetyFactor = 0.90;
        real _scalingFactor = _safetyFactor * pow(abs(_error/_tolerance), real(-0.7/9.0)) * pow(_last_norm_error, real(0.4/9.0));
        _scalingFactor = MAX(_scalingFactor, 1.0/5.0);
        _scalingFactor = MIN(_scalingFactor, 7.0);
        if (_error > _tolerance && _scalingFactor > 1.0) {
          // If our step failed don't try and increase our step size. That would be silly.
          _scalingFactor = _safetyFactor * pow(abs(_error/_tolerance), real(-1.0/9.0));
        }
        _old_step = _step;
        _last_norm_error = pow(_safetyFactor/_scalingFactor*pow(_last_norm_error, real(0.4/9.0)), real(9.0/0.7));
        _step *= _scalingFactor;
      }
      
    } while (_discard);
    
    if (_break_next) {
      if (_next_sample_flag[0]) {
        _mg0_sample();
        _next_sample_counter[0]++;
      }
      if (_next_sample_flag[1]) {
        _mg1_sample();
        _next_sample_counter[1]++;
      }
      if (_next_sample_flag[2]) {
        _mg2_sample();
        _next_sample_counter[2]++;
      }
      if (_next_sample_flag[3]) {
        _mg3_sample();
        _next_sample_counter[3]++;
      }
      if (_next_sample_flag[4])
        _next_sample_flag[5] = true;
      else {
        _break_next = false;
        _t_break_next = _segment1_setup_sampling(_next_sample_flag, _next_sample_counter);
      }
    }
    
    if ( (_t_local + _step)*(1.0 + _EPSILON) > _t_break_next) {
      _break_next = true;
      _LOG(_SAMPLE_LOG_LEVEL, "Current timestep: %e\n", _old_step);
      _step = _t_break_next - _t_local;
    }
  } while (!_next_sample_flag[5]);
  
  _SEGMENT1_END:;
  
  _LOG(_SEGMENT_LOG_LEVEL, "Segment 1: minimum timestep: %e maximum timestep: %e\n", _min_step, _max_step);
  _LOG(_SEGMENT_LOG_LEVEL, "  Attempted %li steps, %.2f%% steps failed.\n", _attempted_steps, (100.0*_unsuccessful_steps)/_attempted_steps);
}


inline void _segment1_calculate_delta_a(real _step)
{
  
  
  // Delta A propagation operator for field n
  _segment1_n_operators_evaluate_operator0(_step);
  
}


inline void _segment1_ip_evolve(int _exponent)
{
}


inline void _segment1_calculate_nonconstant_ip_fields(real _step, int _exponent)
{
}
real _segment1_setup_sampling(bool* _next_sample_flag, long* _next_sample_counter)
{
  // The numbers of the moment groups that need to be sampled at the next sampling point.
  // An entry of N+1 means "reached end of integration interval"
  long _momentGroupNumbersNeedingSamplingNext[5];
  long _numberOfMomentGroupsToBeSampledNext = 1;
  
  long _previous_m = 1;
  long _previous_M = 1;
  
  real _t_break_next = (real)20;
  _momentGroupNumbersNeedingSamplingNext[0] = 4;
  
  // initialise all flags to false
  for (long _i0 = 0; _i0 < 5; _i0++)
    _next_sample_flag[_i0] = false;
  
  /* Check if moment group needs sampling at the same time as another already discovered sample (or the final time).
   * If so, add this moment group to the to-be-sampled list. If moment group demands sampling earlier than all
   * previously noted moment groups, erase all previous ones from list and set the sample time to this earlier one.
   */
  if (_next_sample_counter[0] * _previous_M == _previous_m * 200) {
    _momentGroupNumbersNeedingSamplingNext[_numberOfMomentGroupsToBeSampledNext] = 0;
    _numberOfMomentGroupsToBeSampledNext++;
  } else if (_next_sample_counter[0] * _previous_M < _previous_m * 200) {
    _t_break_next = _next_sample_counter[0] * ((real)20) / ((real)200);
    _numberOfMomentGroupsToBeSampledNext = 1;
    _momentGroupNumbersNeedingSamplingNext[0] = 0;
    _previous_M = 200;
    _previous_m = _next_sample_counter[0];
  }
  
  if (_next_sample_counter[1] * _previous_M == _previous_m * 200) {
    _momentGroupNumbersNeedingSamplingNext[_numberOfMomentGroupsToBeSampledNext] = 1;
    _numberOfMomentGroupsToBeSampledNext++;
  } else if (_next_sample_counter[1] * _previous_M < _previous_m * 200) {
    _t_break_next = _next_sample_counter[1] * ((real)20) / ((real)200);
    _numberOfMomentGroupsToBeSampledNext = 1;
    _momentGroupNumbersNeedingSamplingNext[0] = 1;
    _previous_M = 200;
    _previous_m = _next_sample_counter[1];
  }
  
  if (_next_sample_counter[2] * _previous_M == _previous_m * 200) {
    _momentGroupNumbersNeedingSamplingNext[_numberOfMomentGroupsToBeSampledNext] = 2;
    _numberOfMomentGroupsToBeSampledNext++;
  } else if (_next_sample_counter[2] * _previous_M < _previous_m * 200) {
    _t_break_next = _next_sample_counter[2] * ((real)20) / ((real)200);
    _numberOfMomentGroupsToBeSampledNext = 1;
    _momentGroupNumbersNeedingSamplingNext[0] = 2;
    _previous_M = 200;
    _previous_m = _next_sample_counter[2];
  }
  
  if (_next_sample_counter[3] * _previous_M == _previous_m * 200) {
    _momentGroupNumbersNeedingSamplingNext[_numberOfMomentGroupsToBeSampledNext] = 3;
    _numberOfMomentGroupsToBeSampledNext++;
  } else if (_next_sample_counter[3] * _previous_M < _previous_m * 200) {
    _t_break_next = _next_sample_counter[3] * ((real)20) / ((real)200);
    _numberOfMomentGroupsToBeSampledNext = 1;
    _momentGroupNumbersNeedingSamplingNext[0] = 3;
    _previous_M = 200;
    _previous_m = _next_sample_counter[3];
  }
  
  // _momentGroupNumbersNeedingSamplingNext now contains the complete list of moment groups that need
  // to be sampled at the next sampling point. Set their flags to true.
  for (long _i0 = 0; _i0 < _numberOfMomentGroupsToBeSampledNext; _i0++)
    _next_sample_flag[_momentGroupNumbersNeedingSamplingNext[_i0]] = true;
  
  return _t_break_next;
}

real _segment1_n_coefficients_timestep_error(complex* _checkfield)
{
  real _error = 1e-24;
  real _temp_error = 0.0;
  real _temp_mod = 0.0;

  
  // Find the peak value for each component of the field
  real _cutoff[_n_coefficients_ncomponents];
  
  for (long _i0 = 0; _i0 < _n_coefficients_ncomponents; _i0++)
    _cutoff[_i0] = 0.0;
  
  {
    long _n_coefficients_index_pointer = 0;
    #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
    #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
    for (long _i0 = 0; _i0 < (_lattice_n); _i0++) {
      for (long _i1 = 0; _i1 < _n_coefficients_ncomponents; _i1++) {
        _temp_mod = mod2(_n_coefficients[_n_coefficients_index_pointer + _i1]);
        if (_xmds_isnonfinite(_temp_mod))
          _cutoff[_i1] = INFINITY;
        else if (_cutoff[_i1] < _temp_mod)
          _cutoff[_i1] = _temp_mod;
      }
    
      _n_coefficients_index_pointer += _n_coefficients_ncomponents;
    }
    #undef cup
    #undef cdown
  }
  
  for (long _i0 = 0; _i0 < _n_coefficients_ncomponents; _i0++) {
    if (_xmds_isnonfinite(_cutoff[_i0]))
      // Return an error two times the tolerance in this case because the timestep must be reduced.
      return 2.0*1e-08;
    _cutoff[_i0] *= 0.001;
    _cutoff[_i0] *= 0.001;
  }
  
  {
    long _n_coefficients_index_pointer = 0;
    #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
    #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
    for (long _i0 = 0; _i0 < (_lattice_n); _i0++) {
      for (long  _i1 = 0; _i1 < _n_coefficients_ncomponents; _i1++) {
        if (mod2(_n_coefficients[_n_coefficients_index_pointer + _i1]) > _cutoff[_i1]) {
          _temp_error = abs(_n_coefficients[_n_coefficients_index_pointer + _i1] - _checkfield[_n_coefficients_index_pointer + _i1]) / (0.5*abs(_n_coefficients[_n_coefficients_index_pointer + _i1]) + 0.5*abs(_checkfield[_n_coefficients_index_pointer + _i1]));
          
          if (_xmds_isnonfinite(_temp_error)) {
            /* For _temp_error to be NaN, both the absolute value of the higher and lower order solutions
               must BOTH be zero. This therefore implies that their difference is zero, and that there is no error. */
            _temp_error = 0.0;
          }
          
          if (_error < _temp_error) // UNVECTORISABLE
            _error = _temp_error;
        }
      }
    
      _n_coefficients_index_pointer += _n_coefficients_ncomponents;
    }
    #undef cup
    #undef cdown
  }
  
  return _error;
}

bool _segment1_n_coefficients_reset(complex* _reset_to_n_coefficients)
{
  memcpy(_n_coefficients, _reset_to_n_coefficients, sizeof(complex) * _n_coefficients_alloc_size);
  
  /* return false if there's a NaN somewhere in the vector, otherwise return true */
  bool bNoNaNsPresent = true;
  {
    long _n_coefficients_index_pointer = 0;
    #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
    #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
    for (long _i0 = 0; _i0 < (_lattice_n); _i0++) {
        for (long _i1 = 0; _i1 < _n_coefficients_ncomponents; _i1++) {
          if (_xmds_isnonfinite(_n_coefficients[_n_coefficients_index_pointer + _i1].Re())
            || _xmds_isnonfinite(_n_coefficients[_n_coefficients_index_pointer + _i1].Im())) bNoNaNsPresent = false;
        }
    
      _n_coefficients_index_pointer += _n_coefficients_ncomponents;
    }
    #undef cup
    #undef cdown
  }
  return bNoNaNsPresent;
}

// Delta A propagation operator for field n
void _segment1_n_operators_evaluate_operator0(real _step)
{
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  long _segment1_operator0_delta_a_field_coefficients_index_pointer = 0;
  #define dcup_dt _active_segment1_operator0_delta_a_field_coefficients[_segment1_operator0_delta_a_field_coefficients_index_pointer + 0]
  #define dcdown_dt _active_segment1_operator0_delta_a_field_coefficients[_segment1_operator0_delta_a_field_coefficients_index_pointer + 1]
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    
    #define dt _step
    
    // ************* Propagation code ***************
    #define _cdown_n(_index_n) _active_n_coefficients[1 + (0 \
       + _index_n * 1) * _n_coefficients_ncomponents]
    #define _cup_n(_index_n) _active_n_coefficients[0 + (0 \
       + _index_n * 1) * _n_coefficients_ncomponents]
    #line 127 "Assignment1_program_solutionb_v1.xmds"
    
    
    complex coupling_term_up = 0.0;
     if(n-1>=_min_n)
       coupling_term_up += sqrt(n)*_cup_n(/* n => n -1 */ ((n -1) - _min_n));
     if(n+1<= _max_n)
       coupling_term_up += sqrt(n+1)*_cup_n(/* n => n +1 */ ((n +1) - _min_n));  
    
    
    complex coupling_term_down = 0.0;
     if(n-1>=_min_n)
       coupling_term_down += sqrt(n)*_cdown_n(/* n => n -1 */ ((n -1) - _min_n));
     if(n+1<= _max_n)
       coupling_term_down += sqrt(n+1)*_cdown_n(/* n => n +1 */ ((n +1) - _min_n));           
    
    dcup_dt =   -i*(+0.5*omega0 + omega_i*n + chi_i*n*(n-1))*cup  - i*kappa_i_bar*coupling_term_up +i*0.5*Delta0*cdown;
    dcdown_dt = -i*(-0.5*omega0 + omega_i*n + chi_i*n*(n-1))*cdown + i*kappa_i_bar*coupling_term_down +i*0.5*Delta0*cup;         
    
    
    #line 2354 "Assignment1_program_solutionb_v1.cc"
    #undef _cdown_n
    #undef _cup_n
    // **********************************************
    
    #undef dt
    
    
    // Increment index pointers for vectors in field segment1_operator0_looping_field (or having the same dimensions)
    _n_coefficients_index_pointer += 1 * _n_coefficients_ncomponents;
    _segment1_operator0_delta_a_field_coefficients_index_pointer += 1 * _segment1_operator0_delta_a_field_coefficients_ncomponents;
    
  }
  #undef n
  #undef dn
  _segment1_n_operators_operator0_copy_delta_a(_step);
  #undef cup
  #undef cdown
  #undef dcup_dt
  #undef dcdown_dt
}


inline void _segment1_n_operators_operator0_copy_delta_a(real _step)
{
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  long _segment1_operator0_delta_a_field_coefficients_index_pointer = 0;
  #define dcup_dt _active_segment1_operator0_delta_a_field_coefficients[_segment1_operator0_delta_a_field_coefficients_index_pointer + 0]
  #define dcdown_dt _active_segment1_operator0_delta_a_field_coefficients[_segment1_operator0_delta_a_field_coefficients_index_pointer + 1]
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    
    // UNVECTORISABLE
    // This code copies the increments for the components back into the vectors themselves.
    cup = dcup_dt * _step;
    cdown = dcdown_dt * _step;
    // Increment index pointers for vectors in field segment1_operator0_looping_field (or having the same dimensions)
    _n_coefficients_index_pointer += 1 * _n_coefficients_ncomponents;
    _segment1_operator0_delta_a_field_coefficients_index_pointer += 1 * _segment1_operator0_delta_a_field_coefficients_ncomponents;
    
  }
  #undef n
  #undef dn
  #undef cup
  #undef cdown
  #undef dcup_dt
  #undef dcdown_dt
}

// ********************************************************
//   output function implementations
void _write_output()
{
  _LOG(_SIMULATION_LOG_LEVEL, "Generating output for Assignment1_program_solutionb_v1\n");
  
  // If we aren't rank 0, then we don't want to write anything.
  if (_rank != 0)
    return;
  
  char *_xsilFilename = (char*)malloc(256);
  snprintf(_xsilFilename, 256, "%s.xsil", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
  
  FILE* _outfile = _open_xsil_file(_xsilFilename);
  
  if (_outfile) {
    _write_xsil_header(_outfile);
    char _dataFilename[200];
    snprintf(_dataFilename, 200, "%s.h5", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
    
    H5Fclose(H5Fcreate(_dataFilename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT));
  }
  _mg0_write_out(_outfile);
  _mg1_write_out(_outfile);
  _mg2_write_out(_outfile);
  _mg3_write_out(_outfile);
  
  _write_xsil_footer(_outfile);
  _close_xsil_file(_outfile);
  free(_xsilFilename);
  _xsilFilename = NULL;
  _outfile = NULL;
  
}


FILE* _open_xsil_file(const char* _filename)
{
  
  FILE* fp = fopen(_filename, "w");
  
  if (fp == NULL)
    // _LOG will cause the simulation to exit
    _LOG(_ERROR_LOG_LEVEL, "Unable to open output file '%s'.\n"
                           "Exiting.\n", _filename);
  
  return fp;
}

void _close_xsil_file(FILE*& fp)
{
  if (fp)
    fclose(fp);
  fp = NULL;
  
}

void _write_xsil_header(FILE* fp)
{
  if (!fp)
    return;
  fprintf(fp, "<?xml version=\"1.0\" ?><!DOCTYPE simulation [\n");
  fprintf(fp, "<!ENTITY Nstates     \"21\">\n");
  fprintf(fp, "<!ENTITY nmin    \"0\">\n");
  fprintf(fp, "<!ENTITY nmax    \"20\">\n");
  fprintf(fp, "<!ENTITY Npts     \"256\">\n");
  fprintf(fp, "<!ENTITY L    \"8\">\n");
  fprintf(fp, "]><simulation xmds-version=\"2\">\n");
  fprintf(fp, "  <name>Assignment1_program_solutionb_v1</name>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <author> Sebastian Wuester (Jul 2018)</author>\n");
  fprintf(fp, "  <description>\n");
  fprintf(fp, "    Numerical solution of spin-1/2 + oscillator model for Assignment 1\n");
  fprintf(fp, "  </description>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <geometry>\n");
  fprintf(fp, "    <propagation_dimension> t </propagation_dimension>\n");
  fprintf(fp, "    <transverse_dimensions>\n");
  fprintf(fp, "      <dimension name=\"n\" type=\"integer\" lattice=\"21\" domain=\"(0,20)\" aliases=\"nprime\"/>  <!--Oscillator  basis -->\n");
  fprintf(fp, "      <dimension name=\"x\" lattice=\"256\" domain=\"(-8,8)\"/>  \n");
  fprintf(fp, "     </transverse_dimensions>\n");
  fprintf(fp, "  </geometry>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "    <driver name=\"mpi-multi-path\" paths=\"1\"/>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <features>\n");
  fprintf(fp, "    <benchmark/>\n");
  fprintf(fp, "    <auto_vectorise/>\n");
  fprintf(fp, "    <fftw/>\n");
  fprintf(fp, "    <globals>\n");
  fprintf(fp, "      <![CDATA[\n");
  fprintf(fp, "       // Parameters, some dimensionless units, where hbar=1, sigmaosc=1\n");
  fprintf(fp, "       const double omega0 = 0.0;\n");
  fprintf(fp, "       const double Delta0 = (2.0*M_PI)*0.3;\n");
  fprintf(fp, "       \n");
  fprintf(fp, "       const double omega_i = (2.0*M_PI)*0.4;\n");
  fprintf(fp, "       const double chi_i = (2.0*M_PI)*0.2;\n");
  fprintf(fp, "       const double kappa_i_bar = (2.0*M_PI)*0.15;     \n");
  fprintf(fp, "       \n");
  fprintf(fp, "       const double sigmaosc = 1.0;  \n");
  fprintf(fp, "      \n");
  fprintf(fp, "       //initial state   \n");
  fprintf(fp, "       const double c1 = 1.0;                               //for (b)\n");
  fprintf(fp, "       const double c2 = 0.0;    \n");
  fprintf(fp, " \n");
  fprintf(fp, "       //***************************************************************  \n");
  fprintf(fp, "       // FUNCTION LIBRARY (room for own, non-xmds function definitions)      \n");
  fprintf(fp, "       //***************************************************************  \n");
  fprintf(fp, "       unsigned long factorial(unsigned long f){\n");
  fprintf(fp, "          if (f) return(f * factorial(f - 1));\n");
  fprintf(fp, "          return 1;\n");
  fprintf(fp, "       }\n");
  fprintf(fp, "\n");
  fprintf(fp, "       //***************************************************************\n");
  fprintf(fp, "       unsigned long integerpower(unsigned long argument,unsigned long  exponent){\n");
  fprintf(fp, "       unsigned long result=1;\n");
  fprintf(fp, "     \n");
  fprintf(fp, "       for(unsigned long mm=0;mm<exponent;mm++)     \n");
  fprintf(fp, "          result*=argument;\n");
  fprintf(fp, "         \n");
  fprintf(fp, "          return(result);    \n");
  fprintf(fp, "       }\n");
  fprintf(fp, "\n");
  fprintf(fp, "       //*************************************************************** \n");
  fprintf(fp, "       double Hermite(double Rj,unsigned long degree){\n");
  fprintf(fp, "         if(degree==0) \n");
  fprintf(fp, "           return 1.00000000000000000000;\n");
  fprintf(fp, "        else if(degree==1)\n");
  fprintf(fp, "           return 2.000000000000000*Rj;\n");
  fprintf(fp, "        else{ \n");
  fprintf(fp, "           unsigned long alpha=degree-1;\n");
  fprintf(fp, "           double value=2.000000000000000*(Rj*Hermite(Rj,alpha)-alpha*Hermite(Rj,alpha-1));\n");
  fprintf(fp, "           return value;\n");
  fprintf(fp, "        }      \n");
  fprintf(fp, "       }   \n");
  fprintf(fp, "           \n");
  fprintf(fp, "      ]]>\n");
  fprintf(fp, "    </globals>\n");
  fprintf(fp, "  </features>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <vector name=\"coefficients\" dimensions=\"n\" type=\"complex\">\n");
  fprintf(fp, "    <components>cup cdown</components>\n");
  fprintf(fp, "    <initialisation>\n");
  fprintf(fp, "      <![CDATA[\n");
  fprintf(fp, "        cup = 0.0;\n");
  fprintf(fp, "        cdown = 0.0;\n");
  fprintf(fp, "        \n");
  fprintf(fp, "        if(n==0){//oscillator in the ground-state\n");
  fprintf(fp, "           cup = c1;\n");
  fprintf(fp, "           cdown = c2;\n");
  fprintf(fp, "        }    \n");
  fprintf(fp, "      ]]>\n");
  fprintf(fp, "    </initialisation>\n");
  fprintf(fp, "  </vector>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <vector name=\"oscillator_states\" dimensions=\"n x\" type=\"real\">\n");
  fprintf(fp, "    <components> varphi </components>\n");
  fprintf(fp, "    <initialisation>\n");
  fprintf(fp, "      <![CDATA[\n");
  fprintf(fp, "                double normfact = 1.0/sqrt(sqrt(M_PI)*sigmaosc*factorial(n)*integerpower(2,n));\n");
  fprintf(fp, "                varphi = normfact*exp(-0.5*x*x/sigmaosc/sigmaosc)*Hermite(x/sigmaosc,n);\n");
  fprintf(fp, "      ]]>\n");
  fprintf(fp, "    </initialisation>\n");
  fprintf(fp, "  </vector>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <computed_vector name=\"densities\" dimensions=\"x\" type=\"complex\">\n");
  fprintf(fp, "    <components> phiposup phiposdown </components>\n");
  fprintf(fp, "      <evaluation>\n");
  fprintf(fp, "      <dependencies basis=\"n x\"> coefficients oscillator_states </dependencies> \n");
  fprintf(fp, "      <![CDATA[\n");
  fprintf(fp, "         phiposup = varphi*cup;\n");
  fprintf(fp, "         phiposdown = varphi*cdown;\n");
  fprintf(fp, "      ]]>\n");
  fprintf(fp, "    </evaluation>\n");
  fprintf(fp, "  </computed_vector>\n");
  fprintf(fp, "\n");
  fprintf(fp, "  <sequence>\n");
  fprintf(fp, "    <!-- ARK89 = use Adaptove 8/9th order Runge-Kutta for time stepping -->\n");
  fprintf(fp, "    <!--         it allows a relative error of \"tolerance\" per step  -->\n");
  fprintf(fp, "    <!-- interval = final time of simulation -->\n");
  fprintf(fp, "    <integrate algorithm=\"ARK89\" interval=\"20\" tolerance=\"1e-8\"> \n");
  fprintf(fp, "      <samples>200 200 200 200</samples>\n");
  fprintf(fp, "      <operators>\n");
  fprintf(fp, "        <integration_vectors>coefficients</integration_vectors>\n");
  fprintf(fp, "        <dependencies> </dependencies>       \n");
  fprintf(fp, "        <![CDATA[\n");
  fprintf(fp, "        \n");
  fprintf(fp, "         complex coupling_term_up = 0.0;\n");
  fprintf(fp, "          if(n-1>=_min_n)\n");
  fprintf(fp, "            coupling_term_up += sqrt(n)*cup(n=>n-1);\n");
  fprintf(fp, "          if(n+1<= _max_n)\n");
  fprintf(fp, "            coupling_term_up += sqrt(n+1)*cup(n=>n+1);  \n");
  fprintf(fp, "\n");
  fprintf(fp, "        \n");
  fprintf(fp, "         complex coupling_term_down = 0.0;\n");
  fprintf(fp, "          if(n-1>=_min_n)\n");
  fprintf(fp, "            coupling_term_down += sqrt(n)*cdown(n=>n-1);\n");
  fprintf(fp, "          if(n+1<= _max_n)\n");
  fprintf(fp, "            coupling_term_down += sqrt(n+1)*cdown(n=>n+1);           \n");
  fprintf(fp, "            \n");
  fprintf(fp, "         dcup_dt =   -i*(+0.5*omega0 + omega_i*n + chi_i*n*(n-1))*cup  - i*kappa_i_bar*coupling_term_up +i*0.5*Delta0*cdown;\n");
  fprintf(fp, "         dcdown_dt = -i*(-0.5*omega0 + omega_i*n + chi_i*n*(n-1))*cdown + i*kappa_i_bar*coupling_term_down +i*0.5*Delta0*cup;         \n");
  fprintf(fp, "         \n");
  fprintf(fp, "         ]]>\n");
  fprintf(fp, "      </operators>\n");
  fprintf(fp, "    </integrate>\n");
  fprintf(fp, "  </sequence>\n");
  fprintf(fp, "  \n");
  fprintf(fp, "  <output format=\"hdf5\">\n");
  fprintf(fp, "    <group>\n");
  fprintf(fp, "      <sampling basis=\"n\" initial_sample=\"yes\">\n");
  fprintf(fp, "        <moments> cup_re cup_im cdown_re cdown_im pop_up pop_down   </moments>\n");
  fprintf(fp, "        <dependencies> coefficients </dependencies>\n");
  fprintf(fp, "        <![CDATA[\n");
  fprintf(fp, "          cup_re = cup.Re();\n");
  fprintf(fp, "          cup_im = cup.Im();\n");
  fprintf(fp, "          cdown_re = cdown.Re();\n");
  fprintf(fp, "          cdown_im = cdown.Im();\n");
  fprintf(fp, "          pop_up = mod2(cup);\n");
  fprintf(fp, "          pop_down = mod2(cdown);\n");
  fprintf(fp, "        ]]>\n");
  fprintf(fp, "      </sampling>\n");
  fprintf(fp, "    </group>\n");
  fprintf(fp, "    <group>\n");
  fprintf(fp, "      <sampling basis=\"n(0)\" initial_sample=\"yes\">\n");
  fprintf(fp, "        <moments> norm normup normdown </moments>\n");
  fprintf(fp, "        <dependencies>coefficients</dependencies>\n");
  fprintf(fp, "        <![CDATA[\n");
  fprintf(fp, "          normup = mod2(cup);\n");
  fprintf(fp, "          normdown = mod2(cdown);\n");
  fprintf(fp, "          norm = mod2(cup) + mod2(cdown);\n");
  fprintf(fp, "        ]]>\n");
  fprintf(fp, "      </sampling>\n");
  fprintf(fp, "    </group>\n");
  fprintf(fp, "    <group>\n");
  fprintf(fp, "      <sampling basis=\"n(0)\" initial_sample=\"yes\">\n");
  fprintf(fp, "        <moments> energy </moments>\n");
  fprintf(fp, "        <dependencies>coefficients</dependencies>\n");
  fprintf(fp, "        <![CDATA[\n");
  fprintf(fp, "          energy = 0.0;\n");
  fprintf(fp, "          complex cupstar = cup.Re() - i*cup.Im();\n");
  fprintf(fp, "          complex cdownstar = cdown.Re() - i*cdown.Im();\n");
  fprintf(fp, "          double pop = mod2(cup) + mod2(cdown); \n");
  fprintf(fp, "          \n");
  fprintf(fp, "          energy +=0.5*omega0*(mod2(cup) - mod2(cdown)) - 0.5*Delta0*(cupstar*cdown + cdownstar*cup).Re(); \n");
  fprintf(fp, "          energy +=(omega_i*n + chi_i*n*(n-1))*pop;\n");
  fprintf(fp, "\n");
  fprintf(fp, "          if(n+1<= _max_n){\n");
  fprintf(fp, "            energy += kappa_i_bar*sqrt(n+1)*(cupstar*cup(n=>n+1)).Re(); \n");
  fprintf(fp, "            energy -= kappa_i_bar*sqrt(n+1)*(cdownstar*cdown(n=>n+1)).Re();        \n");
  fprintf(fp, "          }\n");
  fprintf(fp, "\n");
  fprintf(fp, "          if(n-1>=_min_n){\n");
  fprintf(fp, "            energy += kappa_i_bar*sqrt(n)*(cupstar*cup(n=>n-1)).Re(); \n");
  fprintf(fp, "            energy -= kappa_i_bar*sqrt(n)*(cdownstar*cdown(n=>n-1)).Re();         \n");
  fprintf(fp, "          } \n");
  fprintf(fp, "            \n");
  fprintf(fp, "        ]]>\n");
  fprintf(fp, "      </sampling>\n");
  fprintf(fp, "    </group>\n");
  fprintf(fp, "    <group>\n");
  fprintf(fp, "      <sampling basis=\"x\" initial_sample=\"yes\">\n");
  fprintf(fp, "        <moments> totaldensity density_if_up density_if_down </moments>\n");
  fprintf(fp, "        <dependencies>densities</dependencies>\n");
  fprintf(fp, "        <![CDATA[\n");
  fprintf(fp, "          density_if_up = mod2(phiposup);\n");
  fprintf(fp, "          density_if_down = mod2(phiposdown);\n");
  fprintf(fp, "          totaldensity = density_if_up + density_if_down;\n");
  fprintf(fp, "            \n");
  fprintf(fp, "        ]]>\n");
  fprintf(fp, "      </sampling>\n");
  fprintf(fp, "    </group>\n");
  fprintf(fp, "  </output>\n");
  
  fprintf(fp, "\n<info>\n");
  fprintf(fp, "Script compiled with XMDS2 version 3.1.0 \"My kingdom for a tentacle\" (r)\n");
  fprintf(fp, "See http://www.xmds.org for more information.\n");
  fprintf(fp, "</info>\n");
  
}

// In addition to writing the footer (if 'fp' is not NULL)
// this function closes the fp file pointer.
void _write_xsil_footer(FILE* fp)
{
  if (fp) {
    fprintf(fp, "</simulation>\n");
  }
}

// ********************************************************
//   moment group 0 function implementations
void _mg0_sample()
{
  
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  long _mg0_output_raw_index_pointer = 0;
  #define cup_re _active_mg0_output_raw[_mg0_output_raw_index_pointer + 0]
  #define cup_im _active_mg0_output_raw[_mg0_output_raw_index_pointer + 1]
  #define cdown_re _active_mg0_output_raw[_mg0_output_raw_index_pointer + 2]
  #define cdown_im _active_mg0_output_raw[_mg0_output_raw_index_pointer + 3]
  #define pop_up _active_mg0_output_raw[_mg0_output_raw_index_pointer + 4]
  #define pop_down _active_mg0_output_raw[_mg0_output_raw_index_pointer + 5]
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    // Set index pointers explicitly for (some) vectors
    _mg0_output_raw_index_pointer = ( 0
       + _mg0_output_index_t  * _lattice_n
       + _index_n * 1 ) * _mg0_output_raw_ncomponents;
    #define _SAMPLE_COMPLEX(variable) \
              variable ## R = variable.Re(); variable ## I = variable.Im();
    
    // *************** Sampling code ****************
    #line 155 "Assignment1_program_solutionb_v1.xmds"
    
    cup_re = cup.Re();
    cup_im = cup.Im();
    cdown_re = cdown.Re();
    cdown_im = cdown.Im();
    pop_up = mod2(cup);
    pop_down = mod2(cdown);
    
    #line 2733 "Assignment1_program_solutionb_v1.cc"
    // **********************************************
    
    #undef _SAMPLE_COMPLEX
    // Increment index pointers for vectors in field mg0_sampling (or having the same dimensions)
    _n_coefficients_index_pointer += 1 * _n_coefficients_ncomponents;
    
  }
  #undef n
  #undef dn
  #undef cup
  #undef cdown
  #undef cup_re
  #undef cup_im
  #undef cdown_re
  #undef cdown_im
  #undef pop_up
  #undef pop_down
  
  _mg0_output_t[0 + _mg0_output_index_t++] = t;
  
  _LOG(_SAMPLE_LOG_LEVEL, "Sampled field (for moment group #1) at t = %e\n", t);
  
}


void _mg0_process()
{
  {
    _MAKE_AUTOVEC_VARIABLE(_active_mg0_output_processed);
    _MAKE_AUTOVEC_VARIABLE(_mg0_output_sd);
    _MAKE_AUTOVEC_VARIABLE(_active_mg0_output_raw);
    #pragma ivdep
    for (long _i0 = 0; _i0 < (_mg0_output_lattice_t * _lattice_n) * _mg0_output_raw_ncomponents; _i0++) {
      _AUTOVEC(_active_mg0_output_processed)[_i0] += _AUTOVEC(_active_mg0_output_raw)[_i0];
      _AUTOVEC(_mg0_output_sd)[_i0] += _AUTOVEC(_active_mg0_output_raw)[_i0] * _AUTOVEC(_active_mg0_output_raw)[_i0];
    }
  
  }
}


void _mg0_write_out(FILE* _outfile)
{
  #pragma novector
  for (long _i0 = 0; _i0 < (_mg0_output_lattice_t * _lattice_n) * _mg0_output_processed_ncomponents; _i0++) {
    
    // Calculate the mean
    _active_mg0_output_processed[_i0] /= (real) _n_paths;
    
    // Calculate the standard error
    _mg0_output_sd[_i0] /= (real) _n_paths;
    _mg0_output_sd[_i0] -= _active_mg0_output_processed[_i0] * _active_mg0_output_processed[_i0];
    if (_mg0_output_sd[_i0] > 0.0) // UNVECTORISABLE
      _mg0_output_sd[_i0] = sqrt(_mg0_output_sd[_i0] / _n_paths);
    else
      _mg0_output_sd[_i0] = 0.0;
  }
  
  
  if (_outfile) {
    fprintf(_outfile, "\n");
    fprintf(_outfile, "<XSIL Name=\"moment_group_1\">\n");
    fprintf(_outfile, "  <Param Name=\"n_independent\">2</Param>\n");
    fprintf(_outfile, "  <Array Name=\"variables\" Type=\"Text\">\n");
    fprintf(_outfile, "    <Dim>14</Dim>\n");
    fprintf(_outfile, "    <Stream><Metalink Format=\"Text\" Delimiter=\" \\n\"/>\n");
    fprintf(_outfile, "t n mean_cup_re mean_cup_im mean_cdown_re mean_cdown_im mean_pop_up mean_pop_down stderr_cup_re stderr_cup_im stderr_cdown_re stderr_cdown_im stderr_pop_up stderr_pop_down \n");
    fprintf(_outfile, "    </Stream>\n");
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "  <Array Name=\"data\" Type=\"double\">\n");
    fprintf(_outfile, "    <Dim>%i</Dim>\n", _mg0_output_lattice_t);
    fprintf(_outfile, "    <Dim>%i</Dim>\n", _lattice_n);
    fprintf(_outfile, "    <Dim>14</Dim>\n");
  }
  
  
  char _h5Filename[200];
  snprintf(_h5Filename, 200, "%s.h5", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
  
  /* Open the file */
  hid_t hdf5_file = H5Fopen(_h5Filename, H5F_ACC_RDWR, H5P_DEFAULT);
  if (hdf5_file < 0) {
    _LOG(_WARNING_LOG_LEVEL, "Failed to open HDF5 file '%s', will try to create it.", _h5Filename);
    hdf5_file = H5Fcreate(_h5Filename, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT);
    if (hdf5_file < 0) {
      _LOG(_ERROR_LOG_LEVEL, "Failed to create HDF5 file '%s'. Bailing.", _h5Filename);
    }
  }
  
  /* Create the group for this data */
  hid_t group;
  if (!H5Lexists(hdf5_file, "/1", H5P_DEFAULT))
    group = H5Gcreate(hdf5_file, "/1", H5P_DEFAULT);
  else
    group = H5Gopen(hdf5_file, "/1");
  
  if (_outfile) {
    fprintf(_outfile, "    <Stream><Metalink Format=\"HDF5\" Type=\"Remote\" Group=\"/1\"/>\n");
    fprintf(_outfile, "%s.h5\n", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
    fprintf(_outfile, "    </Stream>\n");
  }
  
  /* Create the coordinate data sets */
  hsize_t coordinate_length;
  hid_t coordinate_dataspace;
  coordinate_length = _mg0_output_lattice_t;
  coordinate_dataspace = H5Screate_simple(1, &coordinate_length, NULL);
  hid_t dataset_t;
  if (!H5Lexists(hdf5_file, "/1/t", H5P_DEFAULT))
    dataset_t = H5Dcreate(hdf5_file, "/1/t", H5T_NATIVE_REAL, coordinate_dataspace, H5P_DEFAULT);
  else
    dataset_t = H5Dopen(hdf5_file, "/1/t");
  H5Dwrite(dataset_t, H5T_NATIVE_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT, _mg0_output_t);
  #if defined(HAVE_HDF5_HL)
    H5DSset_scale(dataset_t, "t");
  #endif
  
  H5Sclose(coordinate_dataspace);
  coordinate_length = _lattice_n;
  coordinate_dataspace = H5Screate_simple(1, &coordinate_length, NULL);
  hid_t dataset_n;
  if (!H5Lexists(hdf5_file, "/1/n", H5P_DEFAULT))
    dataset_n = H5Dcreate(hdf5_file, "/1/n", H5T_NATIVE_LONG, coordinate_dataspace, H5P_DEFAULT);
  else
    dataset_n = H5Dopen(hdf5_file, "/1/n");
  H5Dwrite(dataset_n, H5T_NATIVE_LONG, H5S_ALL, H5S_ALL, H5P_DEFAULT, _n);
  #if defined(HAVE_HDF5_HL)
    H5DSset_scale(dataset_n, "n");
  #endif
  
  H5Sclose(coordinate_dataspace);
  
  hsize_t file_dims[] = {_mg0_output_lattice_t, _lattice_n};
  hid_t file_dataspace = H5Screate_simple(2, file_dims, NULL);
  
  hid_t dataset_mean_cup_re;
  if (!H5Lexists(hdf5_file, "/1/mean_cup_re", H5P_DEFAULT))
    dataset_mean_cup_re = H5Dcreate(hdf5_file, "/1/mean_cup_re", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_cup_re = H5Dopen(hdf5_file, "/1/mean_cup_re");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_cup_re, dataset_t, 0);
    H5DSattach_scale(dataset_mean_cup_re, dataset_n, 1);
  #endif
  hid_t dataset_mean_cup_im;
  if (!H5Lexists(hdf5_file, "/1/mean_cup_im", H5P_DEFAULT))
    dataset_mean_cup_im = H5Dcreate(hdf5_file, "/1/mean_cup_im", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_cup_im = H5Dopen(hdf5_file, "/1/mean_cup_im");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_cup_im, dataset_t, 0);
    H5DSattach_scale(dataset_mean_cup_im, dataset_n, 1);
  #endif
  hid_t dataset_mean_cdown_re;
  if (!H5Lexists(hdf5_file, "/1/mean_cdown_re", H5P_DEFAULT))
    dataset_mean_cdown_re = H5Dcreate(hdf5_file, "/1/mean_cdown_re", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_cdown_re = H5Dopen(hdf5_file, "/1/mean_cdown_re");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_cdown_re, dataset_t, 0);
    H5DSattach_scale(dataset_mean_cdown_re, dataset_n, 1);
  #endif
  hid_t dataset_mean_cdown_im;
  if (!H5Lexists(hdf5_file, "/1/mean_cdown_im", H5P_DEFAULT))
    dataset_mean_cdown_im = H5Dcreate(hdf5_file, "/1/mean_cdown_im", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_cdown_im = H5Dopen(hdf5_file, "/1/mean_cdown_im");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_cdown_im, dataset_t, 0);
    H5DSattach_scale(dataset_mean_cdown_im, dataset_n, 1);
  #endif
  hid_t dataset_mean_pop_up;
  if (!H5Lexists(hdf5_file, "/1/mean_pop_up", H5P_DEFAULT))
    dataset_mean_pop_up = H5Dcreate(hdf5_file, "/1/mean_pop_up", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_pop_up = H5Dopen(hdf5_file, "/1/mean_pop_up");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_pop_up, dataset_t, 0);
    H5DSattach_scale(dataset_mean_pop_up, dataset_n, 1);
  #endif
  hid_t dataset_mean_pop_down;
  if (!H5Lexists(hdf5_file, "/1/mean_pop_down", H5P_DEFAULT))
    dataset_mean_pop_down = H5Dcreate(hdf5_file, "/1/mean_pop_down", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_pop_down = H5Dopen(hdf5_file, "/1/mean_pop_down");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_pop_down, dataset_t, 0);
    H5DSattach_scale(dataset_mean_pop_down, dataset_n, 1);
  #endif
  hid_t dataset_stderr_cup_re;
  if (!H5Lexists(hdf5_file, "/1/stderr_cup_re", H5P_DEFAULT))
    dataset_stderr_cup_re = H5Dcreate(hdf5_file, "/1/stderr_cup_re", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_cup_re = H5Dopen(hdf5_file, "/1/stderr_cup_re");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_cup_re, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_cup_re, dataset_n, 1);
  #endif
  hid_t dataset_stderr_cup_im;
  if (!H5Lexists(hdf5_file, "/1/stderr_cup_im", H5P_DEFAULT))
    dataset_stderr_cup_im = H5Dcreate(hdf5_file, "/1/stderr_cup_im", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_cup_im = H5Dopen(hdf5_file, "/1/stderr_cup_im");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_cup_im, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_cup_im, dataset_n, 1);
  #endif
  hid_t dataset_stderr_cdown_re;
  if (!H5Lexists(hdf5_file, "/1/stderr_cdown_re", H5P_DEFAULT))
    dataset_stderr_cdown_re = H5Dcreate(hdf5_file, "/1/stderr_cdown_re", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_cdown_re = H5Dopen(hdf5_file, "/1/stderr_cdown_re");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_cdown_re, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_cdown_re, dataset_n, 1);
  #endif
  hid_t dataset_stderr_cdown_im;
  if (!H5Lexists(hdf5_file, "/1/stderr_cdown_im", H5P_DEFAULT))
    dataset_stderr_cdown_im = H5Dcreate(hdf5_file, "/1/stderr_cdown_im", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_cdown_im = H5Dopen(hdf5_file, "/1/stderr_cdown_im");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_cdown_im, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_cdown_im, dataset_n, 1);
  #endif
  hid_t dataset_stderr_pop_up;
  if (!H5Lexists(hdf5_file, "/1/stderr_pop_up", H5P_DEFAULT))
    dataset_stderr_pop_up = H5Dcreate(hdf5_file, "/1/stderr_pop_up", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_pop_up = H5Dopen(hdf5_file, "/1/stderr_pop_up");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_pop_up, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_pop_up, dataset_n, 1);
  #endif
  hid_t dataset_stderr_pop_down;
  if (!H5Lexists(hdf5_file, "/1/stderr_pop_down", H5P_DEFAULT))
    dataset_stderr_pop_down = H5Dcreate(hdf5_file, "/1/stderr_pop_down", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_pop_down = H5Dopen(hdf5_file, "/1/stderr_pop_down");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_pop_down, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_pop_down, dataset_n, 1);
  #endif
  H5Dclose(dataset_t);
  H5Dclose(dataset_n);
  
  
  if ((_mg0_output_lattice_t * _lattice_n)) {
    /* Create the data space */
    hsize_t file_start[2] = {(unsigned long long int)0, (unsigned long long int)0};
    hsize_t mem_dims[3] = {(unsigned long long int)_mg0_output_lattice_t, (unsigned long long int)_lattice_n, (unsigned long long int)1};
    hsize_t mem_start[3] = {0, 0, 0};
    hsize_t mem_stride[3] = {1, 1, 1};
    hsize_t mem_count[3] = {(unsigned long long int)_mg0_output_lattice_t, (unsigned long long int)_lattice_n, (unsigned long long int)1};
    
    
    hid_t mem_dataspace;
    mem_dims[2] = 6;
    mem_dataspace = H5Screate_simple(3, mem_dims, NULL);
    mem_stride[2] = 6;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[2] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_cup_re)
      H5Dwrite(dataset_mean_cup_re, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg0_output_processed);
    mem_start[2] = 1;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_cup_im)
      H5Dwrite(dataset_mean_cup_im, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg0_output_processed);
    mem_start[2] = 2;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_cdown_re)
      H5Dwrite(dataset_mean_cdown_re, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg0_output_processed);
    mem_start[2] = 3;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_cdown_im)
      H5Dwrite(dataset_mean_cdown_im, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg0_output_processed);
    mem_start[2] = 4;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_pop_up)
      H5Dwrite(dataset_mean_pop_up, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg0_output_processed);
    mem_start[2] = 5;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_pop_down)
      H5Dwrite(dataset_mean_pop_down, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg0_output_processed);
    
    H5Sclose(mem_dataspace);
    mem_dims[2] = 6;
    mem_dataspace = H5Screate_simple(3, mem_dims, NULL);
    mem_stride[2] = 6;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[2] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_cup_re)
      H5Dwrite(dataset_stderr_cup_re, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg0_output_sd);
    mem_start[2] = 1;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_cup_im)
      H5Dwrite(dataset_stderr_cup_im, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg0_output_sd);
    mem_start[2] = 2;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_cdown_re)
      H5Dwrite(dataset_stderr_cdown_re, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg0_output_sd);
    mem_start[2] = 3;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_cdown_im)
      H5Dwrite(dataset_stderr_cdown_im, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg0_output_sd);
    mem_start[2] = 4;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_pop_up)
      H5Dwrite(dataset_stderr_pop_up, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg0_output_sd);
    mem_start[2] = 5;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_pop_down)
      H5Dwrite(dataset_stderr_pop_down, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg0_output_sd);
    
    H5Sclose(mem_dataspace);
  }
  
  
  H5Dclose(dataset_mean_cup_re);
  H5Dclose(dataset_mean_cup_im);
  H5Dclose(dataset_mean_cdown_re);
  H5Dclose(dataset_mean_cdown_im);
  H5Dclose(dataset_mean_pop_up);
  H5Dclose(dataset_mean_pop_down);
  H5Dclose(dataset_stderr_cup_re);
  H5Dclose(dataset_stderr_cup_im);
  H5Dclose(dataset_stderr_cdown_re);
  H5Dclose(dataset_stderr_cdown_im);
  H5Dclose(dataset_stderr_pop_up);
  H5Dclose(dataset_stderr_pop_down);
  
  H5Sclose(file_dataspace);
  H5Gclose(group);
  H5Fclose(hdf5_file);
  
  
  if (_outfile) {
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "</XSIL>\n");
  }
}

// ********************************************************
//   field mg0_output function implementations
// initialisation for vector raw
void _mg0_output_raw_initialise()
{
  
  bzero(_active_mg0_output_raw, sizeof(real) * _mg0_output_raw_alloc_size);
}

// initialisation for vector processed
void _mg0_output_processed_initialise()
{
  
  bzero(_active_mg0_output_processed, sizeof(real) * _mg0_output_processed_alloc_size);
}

// ********************************************************
//   moment group 1 function implementations
void _mg1_sample()
{
  
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  long _mg1_output_raw_index_pointer = 0;
  real norm;
  real normup;
  real normdown;
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    // Set index pointers explicitly for (some) vectors
    _mg1_output_raw_index_pointer = ( 0
       + _mg1_output_index_t  * 1 ) * _mg1_output_raw_ncomponents;
    #define _SAMPLE_COMPLEX(variable) \
              variable ## R = variable.Re(); variable ## I = variable.Im();
    
    // *************** Sampling code ****************
    #line 169 "Assignment1_program_solutionb_v1.xmds"
    
    normup = mod2(cup);
    normdown = mod2(cdown);
    norm = mod2(cup) + mod2(cdown);
    
    #line 3149 "Assignment1_program_solutionb_v1.cc"
    // **********************************************
    
    #undef _SAMPLE_COMPLEX
    
    _active_mg1_output_raw[_mg1_output_raw_index_pointer + 0] += norm * dn;
    _active_mg1_output_raw[_mg1_output_raw_index_pointer + 1] += normup * dn;
    _active_mg1_output_raw[_mg1_output_raw_index_pointer + 2] += normdown * dn;
    // Increment index pointers for vectors in field mg1_sampling (or having the same dimensions)
    _n_coefficients_index_pointer += 1 * _n_coefficients_ncomponents;
    
  }
  #undef n
  #undef dn
  #undef cup
  #undef cdown
  
  _mg1_output_t[0 + _mg1_output_index_t++] = t;
  
  _LOG(_SAMPLE_LOG_LEVEL, "Sampled field (for moment group #2) at t = %e\n", t);
  
}


void _mg1_process()
{
  {
    _MAKE_AUTOVEC_VARIABLE(_mg1_output_sd);
    _MAKE_AUTOVEC_VARIABLE(_active_mg1_output_processed);
    _MAKE_AUTOVEC_VARIABLE(_active_mg1_output_raw);
    #pragma ivdep
    for (long _i0 = 0; _i0 < (_mg1_output_lattice_t) * _mg1_output_raw_ncomponents; _i0++) {
      _AUTOVEC(_active_mg1_output_processed)[_i0] += _AUTOVEC(_active_mg1_output_raw)[_i0];
      _AUTOVEC(_mg1_output_sd)[_i0] += _AUTOVEC(_active_mg1_output_raw)[_i0] * _AUTOVEC(_active_mg1_output_raw)[_i0];
    }
  
  }
}


void _mg1_write_out(FILE* _outfile)
{
  #pragma novector
  for (long _i0 = 0; _i0 < (_mg1_output_lattice_t) * _mg1_output_processed_ncomponents; _i0++) {
    
    // Calculate the mean
    _active_mg1_output_processed[_i0] /= (real) _n_paths;
    
    // Calculate the standard error
    _mg1_output_sd[_i0] /= (real) _n_paths;
    _mg1_output_sd[_i0] -= _active_mg1_output_processed[_i0] * _active_mg1_output_processed[_i0];
    if (_mg1_output_sd[_i0] > 0.0) // UNVECTORISABLE
      _mg1_output_sd[_i0] = sqrt(_mg1_output_sd[_i0] / _n_paths);
    else
      _mg1_output_sd[_i0] = 0.0;
  }
  
  
  if (_outfile) {
    fprintf(_outfile, "\n");
    fprintf(_outfile, "<XSIL Name=\"moment_group_2\">\n");
    fprintf(_outfile, "  <Param Name=\"n_independent\">1</Param>\n");
    fprintf(_outfile, "  <Array Name=\"variables\" Type=\"Text\">\n");
    fprintf(_outfile, "    <Dim>7</Dim>\n");
    fprintf(_outfile, "    <Stream><Metalink Format=\"Text\" Delimiter=\" \\n\"/>\n");
    fprintf(_outfile, "t mean_norm mean_normup mean_normdown stderr_norm stderr_normup stderr_normdown \n");
    fprintf(_outfile, "    </Stream>\n");
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "  <Array Name=\"data\" Type=\"double\">\n");
    fprintf(_outfile, "    <Dim>%i</Dim>\n", _mg1_output_lattice_t);
    fprintf(_outfile, "    <Dim>7</Dim>\n");
  }
  
  
  char _h5Filename[200];
  snprintf(_h5Filename, 200, "%s.h5", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
  
  /* Open the file */
  hid_t hdf5_file = H5Fopen(_h5Filename, H5F_ACC_RDWR, H5P_DEFAULT);
  if (hdf5_file < 0) {
    _LOG(_WARNING_LOG_LEVEL, "Failed to open HDF5 file '%s', will try to create it.", _h5Filename);
    hdf5_file = H5Fcreate(_h5Filename, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT);
    if (hdf5_file < 0) {
      _LOG(_ERROR_LOG_LEVEL, "Failed to create HDF5 file '%s'. Bailing.", _h5Filename);
    }
  }
  
  /* Create the group for this data */
  hid_t group;
  if (!H5Lexists(hdf5_file, "/2", H5P_DEFAULT))
    group = H5Gcreate(hdf5_file, "/2", H5P_DEFAULT);
  else
    group = H5Gopen(hdf5_file, "/2");
  
  if (_outfile) {
    fprintf(_outfile, "    <Stream><Metalink Format=\"HDF5\" Type=\"Remote\" Group=\"/2\"/>\n");
    fprintf(_outfile, "%s.h5\n", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
    fprintf(_outfile, "    </Stream>\n");
  }
  
  /* Create the coordinate data sets */
  hsize_t coordinate_length;
  hid_t coordinate_dataspace;
  coordinate_length = _mg1_output_lattice_t;
  coordinate_dataspace = H5Screate_simple(1, &coordinate_length, NULL);
  hid_t dataset_t;
  if (!H5Lexists(hdf5_file, "/2/t", H5P_DEFAULT))
    dataset_t = H5Dcreate(hdf5_file, "/2/t", H5T_NATIVE_REAL, coordinate_dataspace, H5P_DEFAULT);
  else
    dataset_t = H5Dopen(hdf5_file, "/2/t");
  H5Dwrite(dataset_t, H5T_NATIVE_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT, _mg1_output_t);
  #if defined(HAVE_HDF5_HL)
    H5DSset_scale(dataset_t, "t");
  #endif
  
  H5Sclose(coordinate_dataspace);
  
  hsize_t file_dims[] = {_mg1_output_lattice_t};
  hid_t file_dataspace = H5Screate_simple(1, file_dims, NULL);
  
  hid_t dataset_mean_norm;
  if (!H5Lexists(hdf5_file, "/2/mean_norm", H5P_DEFAULT))
    dataset_mean_norm = H5Dcreate(hdf5_file, "/2/mean_norm", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_norm = H5Dopen(hdf5_file, "/2/mean_norm");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_norm, dataset_t, 0);
  #endif
  hid_t dataset_mean_normup;
  if (!H5Lexists(hdf5_file, "/2/mean_normup", H5P_DEFAULT))
    dataset_mean_normup = H5Dcreate(hdf5_file, "/2/mean_normup", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_normup = H5Dopen(hdf5_file, "/2/mean_normup");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_normup, dataset_t, 0);
  #endif
  hid_t dataset_mean_normdown;
  if (!H5Lexists(hdf5_file, "/2/mean_normdown", H5P_DEFAULT))
    dataset_mean_normdown = H5Dcreate(hdf5_file, "/2/mean_normdown", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_normdown = H5Dopen(hdf5_file, "/2/mean_normdown");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_normdown, dataset_t, 0);
  #endif
  hid_t dataset_stderr_norm;
  if (!H5Lexists(hdf5_file, "/2/stderr_norm", H5P_DEFAULT))
    dataset_stderr_norm = H5Dcreate(hdf5_file, "/2/stderr_norm", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_norm = H5Dopen(hdf5_file, "/2/stderr_norm");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_norm, dataset_t, 0);
  #endif
  hid_t dataset_stderr_normup;
  if (!H5Lexists(hdf5_file, "/2/stderr_normup", H5P_DEFAULT))
    dataset_stderr_normup = H5Dcreate(hdf5_file, "/2/stderr_normup", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_normup = H5Dopen(hdf5_file, "/2/stderr_normup");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_normup, dataset_t, 0);
  #endif
  hid_t dataset_stderr_normdown;
  if (!H5Lexists(hdf5_file, "/2/stderr_normdown", H5P_DEFAULT))
    dataset_stderr_normdown = H5Dcreate(hdf5_file, "/2/stderr_normdown", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_normdown = H5Dopen(hdf5_file, "/2/stderr_normdown");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_normdown, dataset_t, 0);
  #endif
  H5Dclose(dataset_t);
  
  
  if ((_mg1_output_lattice_t)) {
    /* Create the data space */
    hsize_t file_start[1] = {(unsigned long long int)0};
    hsize_t mem_dims[2] = {(unsigned long long int)_mg1_output_lattice_t, (unsigned long long int)1};
    hsize_t mem_start[2] = {0, 0};
    hsize_t mem_stride[2] = {1, 1};
    hsize_t mem_count[2] = {(unsigned long long int)_mg1_output_lattice_t, (unsigned long long int)1};
    
    
    hid_t mem_dataspace;
    mem_dims[1] = 3;
    mem_dataspace = H5Screate_simple(2, mem_dims, NULL);
    mem_stride[1] = 3;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[1] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_norm)
      H5Dwrite(dataset_mean_norm, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg1_output_processed);
    mem_start[1] = 1;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_normup)
      H5Dwrite(dataset_mean_normup, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg1_output_processed);
    mem_start[1] = 2;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_normdown)
      H5Dwrite(dataset_mean_normdown, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg1_output_processed);
    
    H5Sclose(mem_dataspace);
    mem_dims[1] = 3;
    mem_dataspace = H5Screate_simple(2, mem_dims, NULL);
    mem_stride[1] = 3;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[1] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_norm)
      H5Dwrite(dataset_stderr_norm, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg1_output_sd);
    mem_start[1] = 1;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_normup)
      H5Dwrite(dataset_stderr_normup, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg1_output_sd);
    mem_start[1] = 2;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_normdown)
      H5Dwrite(dataset_stderr_normdown, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg1_output_sd);
    
    H5Sclose(mem_dataspace);
  }
  
  
  H5Dclose(dataset_mean_norm);
  H5Dclose(dataset_mean_normup);
  H5Dclose(dataset_mean_normdown);
  H5Dclose(dataset_stderr_norm);
  H5Dclose(dataset_stderr_normup);
  H5Dclose(dataset_stderr_normdown);
  
  H5Sclose(file_dataspace);
  H5Gclose(group);
  H5Fclose(hdf5_file);
  
  
  if (_outfile) {
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "</XSIL>\n");
  }
}

// ********************************************************
//   field mg1_output function implementations
// initialisation for vector raw
void _mg1_output_raw_initialise()
{
  
  bzero(_active_mg1_output_raw, sizeof(real) * _mg1_output_raw_alloc_size);
}

// initialisation for vector processed
void _mg1_output_processed_initialise()
{
  
  bzero(_active_mg1_output_processed, sizeof(real) * _mg1_output_processed_alloc_size);
}

// ********************************************************
//   moment group 2 function implementations
void _mg2_sample()
{
  
  long _n_coefficients_index_pointer = 0;
  #define cup _active_n_coefficients[_n_coefficients_index_pointer + 0]
  #define cdown _active_n_coefficients[_n_coefficients_index_pointer + 1]
  long _mg2_output_raw_index_pointer = 0;
  real energy;
  #define n _n[_index_n + 0]
  #define dn (_dn * (1.0))
  
  for (long _index_n = 0; _index_n < _lattice_n; _index_n++) {
    // Set index pointers explicitly for (some) vectors
    _mg2_output_raw_index_pointer = ( 0
       + _mg2_output_index_t  * 1 ) * _mg2_output_raw_ncomponents;
    #define _SAMPLE_COMPLEX(variable) \
              variable ## R = variable.Re(); variable ## I = variable.Im();
    
    // *************** Sampling code ****************
    #define _cdown_n(_index_n) _active_n_coefficients[1 + (0 \
       + _index_n * 1) * _n_coefficients_ncomponents]
    #define _cup_n(_index_n) _active_n_coefficients[0 + (0 \
       + _index_n * 1) * _n_coefficients_ncomponents]
    #line 180 "Assignment1_program_solutionb_v1.xmds"
    
    energy = 0.0;
    complex cupstar = cup.Re() - i*cup.Im();
    complex cdownstar = cdown.Re() - i*cdown.Im();
    double pop = mod2(cup) + mod2(cdown); 
    
    energy +=0.5*omega0*(mod2(cup) - mod2(cdown)) - 0.5*Delta0*(cupstar*cdown + cdownstar*cup).Re(); 
    energy +=(omega_i*n + chi_i*n*(n-1))*pop;
    
    if(n+1<= _max_n){
      energy += kappa_i_bar*sqrt(n+1)*(cupstar*_cup_n(/* n => n +1 */ ((n +1) - _min_n))).Re(); 
      energy -= kappa_i_bar*sqrt(n+1)*(cdownstar*_cdown_n(/* n => n +1 */ ((n +1) - _min_n))).Re();        
    }
    
    if(n-1>=_min_n){
      energy += kappa_i_bar*sqrt(n)*(cupstar*_cup_n(/* n => n -1 */ ((n -1) - _min_n))).Re(); 
      energy -= kappa_i_bar*sqrt(n)*(cdownstar*_cdown_n(/* n => n -1 */ ((n -1) - _min_n))).Re();         
    } 
    
    
    #line 3463 "Assignment1_program_solutionb_v1.cc"
    #undef _cdown_n
    #undef _cup_n
    // **********************************************
    
    #undef _SAMPLE_COMPLEX
    
    _active_mg2_output_raw[_mg2_output_raw_index_pointer + 0] += energy * dn;
    // Increment index pointers for vectors in field mg2_sampling (or having the same dimensions)
    _n_coefficients_index_pointer += 1 * _n_coefficients_ncomponents;
    
  }
  #undef n
  #undef dn
  #undef cup
  #undef cdown
  
  _mg2_output_t[0 + _mg2_output_index_t++] = t;
  
  _LOG(_SAMPLE_LOG_LEVEL, "Sampled field (for moment group #3) at t = %e\n", t);
  
}


void _mg2_process()
{
  {
    _MAKE_AUTOVEC_VARIABLE(_active_mg2_output_processed);
    _MAKE_AUTOVEC_VARIABLE(_mg2_output_sd);
    _MAKE_AUTOVEC_VARIABLE(_active_mg2_output_raw);
    #pragma ivdep
    for (long _i0 = 0; _i0 < (_mg2_output_lattice_t) * _mg2_output_raw_ncomponents; _i0++) {
      _AUTOVEC(_active_mg2_output_processed)[_i0] += _AUTOVEC(_active_mg2_output_raw)[_i0];
      _AUTOVEC(_mg2_output_sd)[_i0] += _AUTOVEC(_active_mg2_output_raw)[_i0] * _AUTOVEC(_active_mg2_output_raw)[_i0];
    }
  
  }
}


void _mg2_write_out(FILE* _outfile)
{
  #pragma novector
  for (long _i0 = 0; _i0 < (_mg2_output_lattice_t) * _mg2_output_processed_ncomponents; _i0++) {
    
    // Calculate the mean
    _active_mg2_output_processed[_i0] /= (real) _n_paths;
    
    // Calculate the standard error
    _mg2_output_sd[_i0] /= (real) _n_paths;
    _mg2_output_sd[_i0] -= _active_mg2_output_processed[_i0] * _active_mg2_output_processed[_i0];
    if (_mg2_output_sd[_i0] > 0.0) // UNVECTORISABLE
      _mg2_output_sd[_i0] = sqrt(_mg2_output_sd[_i0] / _n_paths);
    else
      _mg2_output_sd[_i0] = 0.0;
  }
  
  
  if (_outfile) {
    fprintf(_outfile, "\n");
    fprintf(_outfile, "<XSIL Name=\"moment_group_3\">\n");
    fprintf(_outfile, "  <Param Name=\"n_independent\">1</Param>\n");
    fprintf(_outfile, "  <Array Name=\"variables\" Type=\"Text\">\n");
    fprintf(_outfile, "    <Dim>3</Dim>\n");
    fprintf(_outfile, "    <Stream><Metalink Format=\"Text\" Delimiter=\" \\n\"/>\n");
    fprintf(_outfile, "t mean_energy stderr_energy \n");
    fprintf(_outfile, "    </Stream>\n");
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "  <Array Name=\"data\" Type=\"double\">\n");
    fprintf(_outfile, "    <Dim>%i</Dim>\n", _mg2_output_lattice_t);
    fprintf(_outfile, "    <Dim>3</Dim>\n");
  }
  
  
  char _h5Filename[200];
  snprintf(_h5Filename, 200, "%s.h5", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
  
  /* Open the file */
  hid_t hdf5_file = H5Fopen(_h5Filename, H5F_ACC_RDWR, H5P_DEFAULT);
  if (hdf5_file < 0) {
    _LOG(_WARNING_LOG_LEVEL, "Failed to open HDF5 file '%s', will try to create it.", _h5Filename);
    hdf5_file = H5Fcreate(_h5Filename, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT);
    if (hdf5_file < 0) {
      _LOG(_ERROR_LOG_LEVEL, "Failed to create HDF5 file '%s'. Bailing.", _h5Filename);
    }
  }
  
  /* Create the group for this data */
  hid_t group;
  if (!H5Lexists(hdf5_file, "/3", H5P_DEFAULT))
    group = H5Gcreate(hdf5_file, "/3", H5P_DEFAULT);
  else
    group = H5Gopen(hdf5_file, "/3");
  
  if (_outfile) {
    fprintf(_outfile, "    <Stream><Metalink Format=\"HDF5\" Type=\"Remote\" Group=\"/3\"/>\n");
    fprintf(_outfile, "%s.h5\n", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
    fprintf(_outfile, "    </Stream>\n");
  }
  
  /* Create the coordinate data sets */
  hsize_t coordinate_length;
  hid_t coordinate_dataspace;
  coordinate_length = _mg2_output_lattice_t;
  coordinate_dataspace = H5Screate_simple(1, &coordinate_length, NULL);
  hid_t dataset_t;
  if (!H5Lexists(hdf5_file, "/3/t", H5P_DEFAULT))
    dataset_t = H5Dcreate(hdf5_file, "/3/t", H5T_NATIVE_REAL, coordinate_dataspace, H5P_DEFAULT);
  else
    dataset_t = H5Dopen(hdf5_file, "/3/t");
  H5Dwrite(dataset_t, H5T_NATIVE_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT, _mg2_output_t);
  #if defined(HAVE_HDF5_HL)
    H5DSset_scale(dataset_t, "t");
  #endif
  
  H5Sclose(coordinate_dataspace);
  
  hsize_t file_dims[] = {_mg2_output_lattice_t};
  hid_t file_dataspace = H5Screate_simple(1, file_dims, NULL);
  
  hid_t dataset_mean_energy;
  if (!H5Lexists(hdf5_file, "/3/mean_energy", H5P_DEFAULT))
    dataset_mean_energy = H5Dcreate(hdf5_file, "/3/mean_energy", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_energy = H5Dopen(hdf5_file, "/3/mean_energy");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_energy, dataset_t, 0);
  #endif
  hid_t dataset_stderr_energy;
  if (!H5Lexists(hdf5_file, "/3/stderr_energy", H5P_DEFAULT))
    dataset_stderr_energy = H5Dcreate(hdf5_file, "/3/stderr_energy", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_energy = H5Dopen(hdf5_file, "/3/stderr_energy");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_energy, dataset_t, 0);
  #endif
  H5Dclose(dataset_t);
  
  
  if ((_mg2_output_lattice_t)) {
    /* Create the data space */
    hsize_t file_start[1] = {(unsigned long long int)0};
    hsize_t mem_dims[2] = {(unsigned long long int)_mg2_output_lattice_t, (unsigned long long int)1};
    hsize_t mem_start[2] = {0, 0};
    hsize_t mem_stride[2] = {1, 1};
    hsize_t mem_count[2] = {(unsigned long long int)_mg2_output_lattice_t, (unsigned long long int)1};
    
    
    hid_t mem_dataspace;
    mem_dims[1] = 1;
    mem_dataspace = H5Screate_simple(2, mem_dims, NULL);
    mem_stride[1] = 1;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[1] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_energy)
      H5Dwrite(dataset_mean_energy, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg2_output_processed);
    
    H5Sclose(mem_dataspace);
    mem_dims[1] = 1;
    mem_dataspace = H5Screate_simple(2, mem_dims, NULL);
    mem_stride[1] = 1;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[1] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_energy)
      H5Dwrite(dataset_stderr_energy, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg2_output_sd);
    
    H5Sclose(mem_dataspace);
  }
  
  
  H5Dclose(dataset_mean_energy);
  H5Dclose(dataset_stderr_energy);
  
  H5Sclose(file_dataspace);
  H5Gclose(group);
  H5Fclose(hdf5_file);
  
  
  if (_outfile) {
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "</XSIL>\n");
  }
}

// ********************************************************
//   field mg2_output function implementations
// initialisation for vector raw
void _mg2_output_raw_initialise()
{
  
  bzero(_active_mg2_output_raw, sizeof(real) * _mg2_output_raw_alloc_size);
}

// initialisation for vector processed
void _mg2_output_processed_initialise()
{
  
  bzero(_active_mg2_output_processed, sizeof(real) * _mg2_output_processed_alloc_size);
}

// ********************************************************
//   moment group 3 function implementations
void _mg3_sample()
{
  _x_densities_evaluate();
  
  long _mg3_output_raw_index_pointer = 0;
  #define totaldensity _active_mg3_output_raw[_mg3_output_raw_index_pointer + 0]
  #define density_if_up _active_mg3_output_raw[_mg3_output_raw_index_pointer + 1]
  #define density_if_down _active_mg3_output_raw[_mg3_output_raw_index_pointer + 2]
  long _x_densities_index_pointer = 0;
  #define phiposup _active_x_densities[_x_densities_index_pointer + 0]
  #define phiposdown _active_x_densities[_x_densities_index_pointer + 1]
  #define x _x[_index_x + 0]
  #define dx (_dx * (1.0))
  
  for (long _index_x = 0; _index_x < _lattice_x; _index_x++) {
    // Set index pointers explicitly for (some) vectors
    _mg3_output_raw_index_pointer = ( 0
       + _mg3_output_index_t  * _lattice_x
       + _index_x * 1 ) * _mg3_output_raw_ncomponents;
    #define _SAMPLE_COMPLEX(variable) \
              variable ## R = variable.Re(); variable ## I = variable.Im();
    
    // *************** Sampling code ****************
    #line 206 "Assignment1_program_solutionb_v1.xmds"
    
    density_if_up = mod2(phiposup);
    density_if_down = mod2(phiposdown);
    totaldensity = density_if_up + density_if_down;
    
    
    #line 3703 "Assignment1_program_solutionb_v1.cc"
    // **********************************************
    
    #undef _SAMPLE_COMPLEX
    // Increment index pointers for vectors in field mg3_sampling (or having the same dimensions)
    _x_densities_index_pointer += 1 * _x_densities_ncomponents;
    
  }
  #undef x
  #undef dx
  #undef totaldensity
  #undef density_if_up
  #undef density_if_down
  #undef phiposup
  #undef phiposdown
  
  _mg3_output_t[0 + _mg3_output_index_t++] = t;
  
  _LOG(_SAMPLE_LOG_LEVEL, "Sampled field (for moment group #4) at t = %e\n", t);
  
}


void _mg3_process()
{
  {
    _MAKE_AUTOVEC_VARIABLE(_mg3_output_sd);
    _MAKE_AUTOVEC_VARIABLE(_active_mg3_output_processed);
    _MAKE_AUTOVEC_VARIABLE(_active_mg3_output_raw);
    #pragma ivdep
    for (long _i0 = 0; _i0 < (_mg3_output_lattice_t * _lattice_x) * _mg3_output_raw_ncomponents; _i0++) {
      _AUTOVEC(_active_mg3_output_processed)[_i0] += _AUTOVEC(_active_mg3_output_raw)[_i0];
      _AUTOVEC(_mg3_output_sd)[_i0] += _AUTOVEC(_active_mg3_output_raw)[_i0] * _AUTOVEC(_active_mg3_output_raw)[_i0];
    }
  
  }
}


void _mg3_write_out(FILE* _outfile)
{
  #pragma novector
  for (long _i0 = 0; _i0 < (_mg3_output_lattice_t * _lattice_x) * _mg3_output_processed_ncomponents; _i0++) {
    
    // Calculate the mean
    _active_mg3_output_processed[_i0] /= (real) _n_paths;
    
    // Calculate the standard error
    _mg3_output_sd[_i0] /= (real) _n_paths;
    _mg3_output_sd[_i0] -= _active_mg3_output_processed[_i0] * _active_mg3_output_processed[_i0];
    if (_mg3_output_sd[_i0] > 0.0) // UNVECTORISABLE
      _mg3_output_sd[_i0] = sqrt(_mg3_output_sd[_i0] / _n_paths);
    else
      _mg3_output_sd[_i0] = 0.0;
  }
  
  
  if (_outfile) {
    fprintf(_outfile, "\n");
    fprintf(_outfile, "<XSIL Name=\"moment_group_4\">\n");
    fprintf(_outfile, "  <Param Name=\"n_independent\">2</Param>\n");
    fprintf(_outfile, "  <Array Name=\"variables\" Type=\"Text\">\n");
    fprintf(_outfile, "    <Dim>8</Dim>\n");
    fprintf(_outfile, "    <Stream><Metalink Format=\"Text\" Delimiter=\" \\n\"/>\n");
    fprintf(_outfile, "t x mean_totaldensity mean_density_if_up mean_density_if_down stderr_totaldensity stderr_density_if_up stderr_density_if_down \n");
    fprintf(_outfile, "    </Stream>\n");
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "  <Array Name=\"data\" Type=\"double\">\n");
    fprintf(_outfile, "    <Dim>%i</Dim>\n", _mg3_output_lattice_t);
    fprintf(_outfile, "    <Dim>%i</Dim>\n", _lattice_x);
    fprintf(_outfile, "    <Dim>8</Dim>\n");
  }
  
  
  char _h5Filename[200];
  snprintf(_h5Filename, 200, "%s.h5", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
  
  /* Open the file */
  hid_t hdf5_file = H5Fopen(_h5Filename, H5F_ACC_RDWR, H5P_DEFAULT);
  if (hdf5_file < 0) {
    _LOG(_WARNING_LOG_LEVEL, "Failed to open HDF5 file '%s', will try to create it.", _h5Filename);
    hdf5_file = H5Fcreate(_h5Filename, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT);
    if (hdf5_file < 0) {
      _LOG(_ERROR_LOG_LEVEL, "Failed to create HDF5 file '%s'. Bailing.", _h5Filename);
    }
  }
  
  /* Create the group for this data */
  hid_t group;
  if (!H5Lexists(hdf5_file, "/4", H5P_DEFAULT))
    group = H5Gcreate(hdf5_file, "/4", H5P_DEFAULT);
  else
    group = H5Gopen(hdf5_file, "/4");
  
  if (_outfile) {
    fprintf(_outfile, "    <Stream><Metalink Format=\"HDF5\" Type=\"Remote\" Group=\"/4\"/>\n");
    fprintf(_outfile, "%s.h5\n", ("Assignment1_program_solutionb_v1" + gsArgsAndValues).c_str());
    fprintf(_outfile, "    </Stream>\n");
  }
  
  /* Create the coordinate data sets */
  hsize_t coordinate_length;
  hid_t coordinate_dataspace;
  coordinate_length = _mg3_output_lattice_t;
  coordinate_dataspace = H5Screate_simple(1, &coordinate_length, NULL);
  hid_t dataset_t;
  if (!H5Lexists(hdf5_file, "/4/t", H5P_DEFAULT))
    dataset_t = H5Dcreate(hdf5_file, "/4/t", H5T_NATIVE_REAL, coordinate_dataspace, H5P_DEFAULT);
  else
    dataset_t = H5Dopen(hdf5_file, "/4/t");
  H5Dwrite(dataset_t, H5T_NATIVE_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT, _mg3_output_t);
  #if defined(HAVE_HDF5_HL)
    H5DSset_scale(dataset_t, "t");
  #endif
  
  H5Sclose(coordinate_dataspace);
  coordinate_length = _lattice_x;
  coordinate_dataspace = H5Screate_simple(1, &coordinate_length, NULL);
  hid_t dataset_x;
  if (!H5Lexists(hdf5_file, "/4/x", H5P_DEFAULT))
    dataset_x = H5Dcreate(hdf5_file, "/4/x", H5T_NATIVE_REAL, coordinate_dataspace, H5P_DEFAULT);
  else
    dataset_x = H5Dopen(hdf5_file, "/4/x");
  H5Dwrite(dataset_x, H5T_NATIVE_REAL, H5S_ALL, H5S_ALL, H5P_DEFAULT, _x);
  #if defined(HAVE_HDF5_HL)
    H5DSset_scale(dataset_x, "x");
  #endif
  
  H5Sclose(coordinate_dataspace);
  
  hsize_t file_dims[] = {_mg3_output_lattice_t, _lattice_x};
  hid_t file_dataspace = H5Screate_simple(2, file_dims, NULL);
  
  hid_t dataset_mean_totaldensity;
  if (!H5Lexists(hdf5_file, "/4/mean_totaldensity", H5P_DEFAULT))
    dataset_mean_totaldensity = H5Dcreate(hdf5_file, "/4/mean_totaldensity", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_totaldensity = H5Dopen(hdf5_file, "/4/mean_totaldensity");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_totaldensity, dataset_t, 0);
    H5DSattach_scale(dataset_mean_totaldensity, dataset_x, 1);
  #endif
  hid_t dataset_mean_density_if_up;
  if (!H5Lexists(hdf5_file, "/4/mean_density_if_up", H5P_DEFAULT))
    dataset_mean_density_if_up = H5Dcreate(hdf5_file, "/4/mean_density_if_up", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_density_if_up = H5Dopen(hdf5_file, "/4/mean_density_if_up");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_density_if_up, dataset_t, 0);
    H5DSattach_scale(dataset_mean_density_if_up, dataset_x, 1);
  #endif
  hid_t dataset_mean_density_if_down;
  if (!H5Lexists(hdf5_file, "/4/mean_density_if_down", H5P_DEFAULT))
    dataset_mean_density_if_down = H5Dcreate(hdf5_file, "/4/mean_density_if_down", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_mean_density_if_down = H5Dopen(hdf5_file, "/4/mean_density_if_down");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_mean_density_if_down, dataset_t, 0);
    H5DSattach_scale(dataset_mean_density_if_down, dataset_x, 1);
  #endif
  hid_t dataset_stderr_totaldensity;
  if (!H5Lexists(hdf5_file, "/4/stderr_totaldensity", H5P_DEFAULT))
    dataset_stderr_totaldensity = H5Dcreate(hdf5_file, "/4/stderr_totaldensity", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_totaldensity = H5Dopen(hdf5_file, "/4/stderr_totaldensity");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_totaldensity, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_totaldensity, dataset_x, 1);
  #endif
  hid_t dataset_stderr_density_if_up;
  if (!H5Lexists(hdf5_file, "/4/stderr_density_if_up", H5P_DEFAULT))
    dataset_stderr_density_if_up = H5Dcreate(hdf5_file, "/4/stderr_density_if_up", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_density_if_up = H5Dopen(hdf5_file, "/4/stderr_density_if_up");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_density_if_up, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_density_if_up, dataset_x, 1);
  #endif
  hid_t dataset_stderr_density_if_down;
  if (!H5Lexists(hdf5_file, "/4/stderr_density_if_down", H5P_DEFAULT))
    dataset_stderr_density_if_down = H5Dcreate(hdf5_file, "/4/stderr_density_if_down", H5T_NATIVE_REAL, file_dataspace, H5P_DEFAULT);
  else
    dataset_stderr_density_if_down = H5Dopen(hdf5_file, "/4/stderr_density_if_down");
  #if defined(HAVE_HDF5_HL)
    H5DSattach_scale(dataset_stderr_density_if_down, dataset_t, 0);
    H5DSattach_scale(dataset_stderr_density_if_down, dataset_x, 1);
  #endif
  H5Dclose(dataset_t);
  H5Dclose(dataset_x);
  
  
  if ((_mg3_output_lattice_t * _lattice_x)) {
    /* Create the data space */
    hsize_t file_start[2] = {(unsigned long long int)0, (unsigned long long int)0};
    hsize_t mem_dims[3] = {(unsigned long long int)_mg3_output_lattice_t, (unsigned long long int)_lattice_x, (unsigned long long int)1};
    hsize_t mem_start[3] = {0, 0, 0};
    hsize_t mem_stride[3] = {1, 1, 1};
    hsize_t mem_count[3] = {(unsigned long long int)_mg3_output_lattice_t, (unsigned long long int)_lattice_x, (unsigned long long int)1};
    
    
    hid_t mem_dataspace;
    mem_dims[2] = 3;
    mem_dataspace = H5Screate_simple(3, mem_dims, NULL);
    mem_stride[2] = 3;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[2] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_totaldensity)
      H5Dwrite(dataset_mean_totaldensity, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg3_output_processed);
    mem_start[2] = 1;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_density_if_up)
      H5Dwrite(dataset_mean_density_if_up, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg3_output_processed);
    mem_start[2] = 2;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_mean_density_if_down)
      H5Dwrite(dataset_mean_density_if_down, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _active_mg3_output_processed);
    
    H5Sclose(mem_dataspace);
    mem_dims[2] = 3;
    mem_dataspace = H5Screate_simple(3, mem_dims, NULL);
    mem_stride[2] = 3;
    
    // Select hyperslabs of memory and file data spaces for data transfer operation
    mem_start[2] = 0;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_totaldensity)
      H5Dwrite(dataset_stderr_totaldensity, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg3_output_sd);
    mem_start[2] = 1;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_density_if_up)
      H5Dwrite(dataset_stderr_density_if_up, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg3_output_sd);
    mem_start[2] = 2;
    H5Sselect_hyperslab(mem_dataspace, H5S_SELECT_SET, mem_start, mem_stride, mem_count, NULL);
    H5Sselect_hyperslab(file_dataspace, H5S_SELECT_SET, file_start, mem_stride, mem_count, NULL);
    
    if (dataset_stderr_density_if_down)
      H5Dwrite(dataset_stderr_density_if_down, H5T_NATIVE_REAL, mem_dataspace, file_dataspace, H5P_DEFAULT, _mg3_output_sd);
    
    H5Sclose(mem_dataspace);
  }
  
  
  H5Dclose(dataset_mean_totaldensity);
  H5Dclose(dataset_mean_density_if_up);
  H5Dclose(dataset_mean_density_if_down);
  H5Dclose(dataset_stderr_totaldensity);
  H5Dclose(dataset_stderr_density_if_up);
  H5Dclose(dataset_stderr_density_if_down);
  
  H5Sclose(file_dataspace);
  H5Gclose(group);
  H5Fclose(hdf5_file);
  
  
  if (_outfile) {
    fprintf(_outfile, "  </Array>\n");
    fprintf(_outfile, "</XSIL>\n");
  }
}

// ********************************************************
//   field mg3_output function implementations
// initialisation for vector raw
void _mg3_output_raw_initialise()
{
  
  bzero(_active_mg3_output_raw, sizeof(real) * _mg3_output_raw_alloc_size);
}

// initialisation for vector processed
void _mg3_output_processed_initialise()
{
  
  bzero(_active_mg3_output_processed, sizeof(real) * _mg3_output_processed_alloc_size);
}

